#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>
#include "vector.h"
#include "syntax.h"
#include "box.h"
#include "field.h"
#include "expr.h"

//****************************************************************************


extern long location_to_index(double, double, double);
extern int  bcidtoint(const char *);
extern void split(long, int &, int &, int &);
extern char in_bounds(double, double, double);

extern  int   xsize, ysize, zsize;
extern  long  Size, xysize;

extern  double *xgrid, *ygrid, *zgrid;
extern  double *xcoord, *ycoord, *zcoord;

extern  double *dxplus, *dxminus, *dxcenter, *dxinv, *dx2inv;
extern  double *dyplus, *dyminus, *dycenter, *dyinv, *dy2inv;
extern  double *dzplus, *dzminus, *dzcenter, *dzinv, *dz2inv;

extern  double *bc_deriv, *bc_coef, *bc_lin;
extern  char   **bc_id;
extern  int    bc_type_num;

extern  double *right, *diag, *sup, *sub;

extern int VERBOSE;

extern int DIMENSIONALITY;
extern int GEOMETRY;
extern const int SPHERICAL;

const double pi = atan(1.0) * 4;

RegionObj *FieldObj::region = 0;

//****************************************************************************
//*          C L A S S   F I E L D :   I M P L E M E N T A T I O N
//****************************************************************************

void FieldObj::allocate(const char *id)
{
ID = StrCpy(id);

bccond = new int[region -> get_surface_num()];

  ptype = new long[size];

  source_x0    = new int[source_num];
  source_nx    = new int[source_num];
  source_y0    = new int[source_num];
  source_ny    = new int[source_num];
  source_z0    = new int[source_num];
  source_nz    = new int[source_num];
  source_xsum  = new double *[source_num];
  source_ysum  = new double *[source_num];
  source_zsum  = new double *[source_num];
}

//****************************************************************************

/*
FieldObj::FieldObj(int sources, double d, double tot, const char *id) : VectorObj(Size)
{
  D = d;       
  Time = 0.0;  
  current = 0;      

  source_num = sources;
  allocate(id);

  *this = (total = tot);

for (int i = 1; i <= source_num; i++)
  {
  source_xsum[i] = 0;
  source_ysum[i] = 0;
  source_zsum[i] = 0;
  }
}
*/

//***************************************************************************

FieldObj::FieldObj(TokenString &params, const char *id) : VectorObj(Size)
{
  source_num = params.token2_count(id, "source");
  allocate(id);

  Time = 0.0;  current = 0; total = 0.0;
  D = params.trail2_double(id, "D");   

  if (params.token2_count(id, "total")) total = params.trail2_double(id, "total");

 if (VERBOSE) {
   fprintf(stderr, "\n### Concentration field %s: diffusion coefficient = %g um^2 / ms\n", ID, D);
   fprintf(stderr, "    Total [%s] = %g uM \n", ID, total); 
 }

  if (params.token2_count(id, "import"))
    { 
    char filename[255];
    params.trail2_pars(id, "import", 1, 's', filename, 'E');
    if (VERBOSE) fprintf(stderr, "  # Importing %s from file %s\n", ID, filename);
    import(filename);
    }
  else *this = total;

if (source_num)
  {
  double x, y, z, sigmax, sigmay, sigmaz;
  if (VERBOSE) fprintf(stderr, "    %s has %d sources:\n", ID, source_num);
  for (int i = 1; i <= source_num; i++)
    {
    x = y = z = 0;
    sigmax = 0.0; sigmay = sigmaz = -1;

    if (DIMENSIONALITY == 1)
       params.trail2_pars(id, "source", i, 'd', &x, 'd', &sigmax, 'E');
    else  if (DIMENSIONALITY == 2)
       params.trail2_pars(id, "source", i, 'd', &x, 'd', &y, 'd', &sigmax, 'd', &sigmay, 'E');
    else
       params.trail2_pars(id, "source", i, 'd', &x, 'd', &y, 'd', &z, 'd', &sigmax, 
                          'd', &sigmay,  'd', &sigmaz, 'E');

    if (sigmay < 0) sigmay = sigmax;  //  by default widths in all directions are equal
    if (sigmaz < 0) sigmaz = sigmay;  //

    set_source(i-1, x, y, z, sigmax, sigmay, sigmaz); 
    }
  }

  int defs  = params.token2_count(id, "bc");
  int boxes = region->get_box_num();

  char bc1[20], bc2[20], bc3[20], bc4[20], bc5[20], bc6[20];

  for (int i=1; i<= boxes; i++)
    {
    if ( i > defs )   // no b.c. specification for box i:
      {
      if (VERBOSE) fprintf(stderr,"    Box #%d: not specified; assuming all boundaries are Neumann\n", i);
      set_bc(i-1, "Neumann");      
      }
    else if ( params.equal( params.token2_index(id, "bc", i) + 1, "all") ) 
      {
      params.trail2_pars(id, "bc", i, 0,0, 's', bc1, 'E');
      if (VERBOSE) fprintf(stderr,"    Box #%d: all boundaries are %s\n", i, bc1);
      set_bc(i-1, bc1);
      }
    else if (DIMENSIONALITY == 1)
      {
      params.trail2_pars(id, "bc", i, 's', bc1, 's', bc2, 'E');
      if (VERBOSE) 
        fprintf(stderr,"    Box #%d: boundary cond's: xmin:%s xmax:%s\n", i, bc1, bc2);
      set_bc(i-1, bc1, bc2, bc2, bc2, bc2, bc2);
      }
    else if (DIMENSIONALITY == 2)
      {
      params.trail2_pars(id, "bc", i, 's', bc1, 's', bc2, 's', bc3, 's', bc4,  'E');
      if (VERBOSE) 
        fprintf(stderr,"    Box #%d: boundary cond's: xmin:%s xmax:%s ymin:%s ymax:%s\n", i, bc1, bc2, bc3, bc4);
      set_bc(i-1, bc1, bc2, bc3, bc4, bc4, bc4);
      }
    else  // DIMENSIONALITY == 3
      {
      params.trail2_pars(id, "bc", i,
          's', bc1, 's', bc2, 's', bc3, 's', bc4, 's', bc5, 's', bc6, 'E');
      if (VERBOSE) 
        fprintf(stderr,"    Box #%d: boundary cond's: xmin:%s xmax:%s ymin:%s ymax:%s zmin:%s zmax:%s\n", 
                i, bc1, bc2, bc3, bc4, bc5, bc6);
      set_bc(i-1, bc1, bc2, bc3, bc4, bc5, bc6);
      }
    }

  init_boundary();
}

//***************************************************************************

FieldObj::FieldObj(const FieldObj &f) : VectorObj(f)
{
  int nx, ny, nz;
  source_num = f.source_num;
  allocate( f.ID );
 
  current = f.current;

  if (source_num)
    {
    for(int i = 0; i < source_num; i++)  
      {
      source_x0[i] = f.source_x0[i];
      nx = source_nx[i] = f.source_nx[i];
      source_y0[i] = f.source_y0[i];
      ny = source_ny[i] = f.source_ny[i];
      source_z0[i] = f.source_z0[i];
      nz = source_nz[i] = f.source_nz[i];
      
      source_xsum[i] = new double[nx];
      source_ysum[i] = new double[ny];
      source_zsum[i] = new double[nz];

      int ii;
      for (ii = 0; ii < nx; ii++) source_xsum[i][ii] = f.source_xsum[i][ii];
      for (ii = 0; ii < ny; ii++) source_ysum[i][ii] = f.source_ysum[i][ii];
      for (ii = 0; ii < nz; ii++) source_zsum[i][ii] = f.source_zsum[i][ii];
      }
    }

  for(long i=0; i< size; i++) ptype[i] = f.ptype[i];
  for(int n=0; n< region->get_surface_num(); n++)  bccond[n] = f.bccond[n];

  D = f.D;
  Time = f.Time;
  total = f.total;
}

//***************************************************************************

FieldObj &FieldObj::operator=(const FieldObj &f)
{
 *((VectorObj *)this) = f;
 Time = f.Time;
 return *this;
}

//***************************************************************************

FieldObj &FieldObj::operator=(const VectorObj &v)
{
*((VectorObj *)this) = v;

//for (long i = 0; i < size; ++i) elem[i] = v[i]; 

return *this;
}

//***************************************************************************

FieldObj &FieldObj::operator=(const double val)
{
*((VectorObj *)this) = val;

//for (long i = 0; i < size; ++i) elem[i] = val; 
return *this;
}

//************************************************************************

FieldObj &FieldObj::operator=(double (*f0)(double, double, double) )
{
long   i = -1;

for (int iz = 0; iz < zsize; iz++)
  for (int iy = 0; iy < ysize; iy++)
    for (int ix = 0; ix < xsize; ix++)
      if ( ptype[++i] & VARY_MASK )
         elem[i] = (*f0)(xcoord[ix], ycoord[iy], zcoord[iz]);
 
return *this;
}


//************************************************************************************
//                   F I E L D   D E S T R U C T O R
//************************************************************************************

  FieldObj::~FieldObj()
    {
    if (VERBOSE > 7) { fprintf(stderr, "> Destructing field %s:",ID);
                       fflush(stderr); } 

    delete [] ptype;
    delete [] bccond; 
 
    for (int i = 0; i < source_num; i++) 
      {
      delete [] source_xsum[i]; delete [] source_ysum[i]; delete [] source_zsum[i];
      }

    delete [] source_x0;   delete [] source_nx;
    delete [] source_y0;   delete [] source_ny;
    delete [] source_z0;   delete [] source_nz; 

    delete [] ID;
    if (VERBOSE > 7) fprintf(stderr," done\n");
  }


//************************************************************************************
//                   F I E L D   I N I T I A L I Z A T I O N
//************************************************************************************


void  FieldObj::set_bc(int boxid, int bcx0, int bcx1, int bcy0, int bcy1, int bcz0, int bcz1)
{
  if (boxid >= region->get_box_num())  {
    printf("*** ERROR: set_bc: boxid [%d] >= region->total [%d] \n",
          boxid, region->get_box_num());
    exit(1);
    }

  if ( (bcx0 > bc_type_num) || (bcx1 > bc_type_num) || (bcy0 > bc_type_num) ||
       (bcy1 > bc_type_num) || (bcz0 > bc_type_num) || (bcz1 > bc_type_num) )  {
    printf("*** ERROR: set_bc: bcid exceeds number of bc types (%d)\n", bc_type_num);
    exit(1);
    }
  
  bccond[boxid*6]   = bcx0;
  bccond[boxid*6+1] = bcx1;
  bccond[boxid*6+2] = bcy0;
  bccond[boxid*6+3] = bcy1;
  bccond[boxid*6+4] = bcz0;
  bccond[boxid*6+5] = bcz1;

}

//******************************************

void  FieldObj::set_bc(int boxid, const char *bcx0, const char *bcx1, const char *bcy0,
                                  const char *bcy1, const char *bcz0, const char *bcz1)
{
  set_bc( boxid, bcidtoint(bcx0), bcidtoint(bcx1), 
                 bcidtoint(bcy0), bcidtoint(bcy1), 
	         bcidtoint(bcz0), bcidtoint(bcz1) );
}

//******************************************

void  FieldObj::set_bc(int boxid, int bc) {
  set_bc( boxid, bc, bc, bc, bc, bc, bc);
}

//******************************************

void  FieldObj::set_bc(int boxid, const char *bcid) {
  set_bc( boxid, bcidtoint(bcid) ); 
}

//************************************************************************************
//                              S E T   S O U R C E
//************************************************************************************

void  FieldObj::set_source(int id, double x, double y, double z, double sigmax, double sigmay, double sigmaz)
{

  if (VERBOSE) {
     fprintf(stderr, "    Setting source #%d at (%g",id+1, x);
     if (DIMENSIONALITY > 1) fprintf(stderr, ",%g",y);
     if (DIMENSIONALITY > 2) fprintf(stderr, ",%g",z);
     fprintf(stderr, "), with half-width of (%g", sigmax);
     if (DIMENSIONALITY > 1) fprintf(stderr, ",%g",sigmay);
     if (DIMENSIONALITY > 2) fprintf(stderr, ",%g",sigmaz);
     fprintf(stderr, ")\n");
  }

  if (id >= source_num)
    fprintf(stderr, "*** ERROR: set source: source id [%d] >= source_num [%d] \n",id,source_num),
    exit(1);

  if ( !in_bounds(x, y, z) )
    fprintf(stderr, "FieldObj::set_source: location (%g,%g,%g)nm out of bounds\n",x*1e3,y*1e3,z*1e3),
    exit(1);

  long ind = location_to_index(x, y, z);
  
  int ix, iy, iz;
  split(ind, ix, iy, iz);
  if (VERBOSE) {
     fprintf(stderr, "    closest node: (%d", ix);
     if (DIMENSIONALITY > 1) fprintf(stderr, ",%d",iy);
     if (DIMENSIONALITY > 2) fprintf(stderr, ",%d",iz);
     fprintf(stderr, ") <--> (%.4g", xcoord[ix]);
     if (DIMENSIONALITY > 1) fprintf(stderr, ",%.4g",ycoord[iy]);
     if (DIMENSIONALITY > 2) fprintf(stderr, ",%.4g",zcoord[iz]);
     fprintf(stderr, ")\n");
  }

  //*************************

  if (VERBOSE > 2) { fprintf(stderr,"      non-zero current on x-nodes "); fflush(stderr); }
  set_source_xyz(id, x, ind, ix, sigmax, xcoord, xgrid, SURF_XMIN, SURF_XMAX, 1,
                 xsize - 1, source_x0, source_nx, source_xsum);


  if (DIMENSIONALITY > 1) {
    if (VERBOSE > 2) { fprintf(stderr,"      non-zero current on y-nodes "); fflush(stderr); }
    set_source_xyz(id, y, ind, iy, sigmay, ycoord, ygrid, SURF_YMIN, SURF_YMAX, xsize,
                 ysize - 1, source_y0, source_ny, source_ysum);
  } else { 
    source_y0[id] = 0; source_ny[id] = 1;
    source_ysum[id] = new double[1]; source_ysum[id][0] = 1.0;
    }

  if (DIMENSIONALITY > 2) {
    if (VERBOSE > 2)  { fprintf(stderr,"      non-zero current on z-nodes "); fflush(stderr); }
    set_source_xyz(id, z, ind, iz, sigmaz, zcoord, zgrid, SURF_ZMIN, SURF_ZMAX, xysize,
                 zsize - 1, source_z0, source_nz, source_zsum);
    }
  else { 
    source_z0[id] = 0; source_nz[id] = 1;
    source_zsum[id] = new double[1]; source_zsum[id][0] = 1.0;
    }
}

//************************************************************************************
//                            S O U R C E   S P R E A D
//************************************************************************************

void  FieldObj::set_source_xyz(int id, double c0, long i0, int ix, double sigma, 
               double *coord, double *grid, int minMask, int maxMask, int incr, int N,
               int *source_coord0, int *interval_num, double **source_weights)
{
if (sigma < 0.1 * grid[ix])  sigma = 0.1 * grid[ix];  // if source is point-like, set half-width at 
                                                      // 1/10th of grid spacing
  double *temp_weights = new double[N+1];
  double check_sum = 0.0;

  bool flag = false;    // to flag the entry into the region of non-zero current

  long i3d = i0;
  int  ileft = ix, iright = ix;
  while ( !( region->point_type(i3d) & minMask) ) { i3d -= incr; ileft --; } 
  i3d = i0;
  while ( !( region->point_type(i3d) & maxMask) ) { i3d += incr; iright ++; }

  for (int ii = ileft; ii <= iright; ii++) 
    {
    double dx0 = 0.5 * ( (ii > ileft)  ? grid[ii-1] : 0.0 );
    double dx1 = 0.5 * ( (ii < iright) ? grid[ii] : 0.0 );

    double x0 = coord[ii] - c0 - dx0;
    double x1 = coord[ii] - c0 + dx1;
    
    double left0 = 2 * (c0 - coord[ileft]) + x0;
    double left1 = 2 * (c0 - coord[ileft]) + x1;

    double right0 = 2 * (coord[iright] - c0) - x1;
    double right1 = 2 * (coord[iright] - c0) - x0;

    temp_weights[ii] = errf( x0, x1, sigma  ) + 
                       errf( left0, left1, sigma ) + 
                       errf( right0, right1, sigma ); 

    // printf("[%g,%g] [%g,%g] [%g,%g] \n",x0,x1,left0,left1,right0,right1);

    check_sum += temp_weights[ii];

    double dv = (dx0 + dx1);  // default volume element

    if (GEOMETRY == 1)    // Spherical
      dv = 4.0 / 3.0 * pi * ( pow( coord[ii] + dx1, 3.0 ) - pow( coord[ii] - dx0, 3.0 ) );

    else if (GEOMETRY == 2)  // Conical
      if (coord == xcoord)
        dv = 2 * pi / 3 * ( pow( coord[ii] + dx1, 3.0 ) - pow( coord[ii] - dx0, 3.0 ) );
      else if (coord == ycoord)
        dv = cos( coord[ii] - dx0 ) - cos( coord[ii] + dx1 );

    temp_weights[ii] /= dv;
   
    if ( temp_weights[ii] > SOURCE_EPS )   // integral between ii-1/2 and ii+1/2 > eps
      {
       if ( !flag )  
	 {
         flag = true;
         source_coord0[id] = ii;
         interval_num[id] = 1;
	 }
       else interval_num[id]++;
       if (VERBOSE > 2) fprintf(stderr, "%d(%.4g) ", ii, coord[ii]*1e3 );
      }
    }

  //*************************

  source_weights[id] = new double[interval_num[id]];

  for (int i = 0; i < interval_num[id]; i++) 
    source_weights[id][i] = temp_weights[i + source_coord0[id]] / check_sum;   // renormalize 

  if (fabs(check_sum - 1) > 1.0e-4) fprintf(stderr," !!! Current integral error=%.2e\n", check_sum - 1.0);

  if (VERBOSE > 2) fprintf(stderr, "\n");

  delete [] temp_weights;
}

//************************************************************************************
//                      O L D    S O U R C E   S P R E A D
//************************************************************************************

double errf(double x0, double x1, double sigma)
{
  //fprintf(stderr,"x0=%g x1=%g sigma=%g\n", x0, x1, sigma);

  double eps = 1.0e-10;

  double pi = atan(1.0) * 4.0;
  double a = 1.0 / ( pow(pi * sigma * sigma, 0.5) );
  double b = -1.0 / (sigma * sigma);

  long   N = 2;
  double resN = 0, res2N = 0, res4N = 0;
  double x, dx, sum;

  do
    {
    resN = res2N;
    N *= 2;
    dx = (x1 - x0) / double(N);
    sum = 0.0;
    for (int n = 1; n < N; n++)
      {
      x = x0 + dx * double(n);
      sum += ((n & 1) + 1) * exp( b * x * x );
      }
    res2N = dx / 3.0 * a * ( 2 * sum + exp( b * x0 * x0 ) + exp( b * x1 * x1 ) );
    res4N = (16.0 * res2N - resN) / 15.0;
    }
  while ( fabs(res4N - res2N) > eps * fabs(res4N + res2N) && fabs(res4N + res2N) > 1e-12);

  return res4N;
}

//***************************************************************

void  FieldObj::add_sources(VectorObj &f, double dt)
{
int    i, ix, iy, iz;
long   xind, yind, zind;
double addx, addy;

 double I = current->Evaluate();
 if (I < MIN_CURRENT) return;
 double charge = I * dt;

  for (i=0; i < source_num; i++) 
    {
    int nx = source_nx[i];
    int ny = source_ny[i];
    int nz = source_nz[i];
    xind = source_x0[i] + xsize * source_y0[i] + xysize * source_z0[i];

    for (ix = 0; ix < nx; ix++, xind ++)
      {
      addx = charge * source_xsum[i][ix];
      for (iy = 0, yind = xind; iy < ny; iy++, yind += xsize)
	{
	addy = addx * source_ysum[i][iy];
        for (iz = 0, zind = yind; iz < nz; iz++, zind += xysize)
           f.elem[zind] += addy * source_zsum[i][iz];

	}
      }
    }
}

//************************************************************************
//            B O U N D A R Y   I N I T I A L I Z A T I O N
//************************************************************************
/*
   RegionObj::point_type(int,int,int) bit fields are:
   bit # is:      ...8       |   7    |   6  |  5   |  4   |  3   |  2   |  1  
   meaning:    box id number | inside | zmax | zmin | ymax | ymin | xmax | xmin

  ***************************************************************

   ptype array bit fields are:

   bit # is:  ...10    |    9    |    8      |   7    |   6  |  5   |  4   |  3   |  2   |  1  
   meaning:  bc_type # | outside | dirichlet | inside | zmax | zmin | ymax | ymin | xmax | xmin
  
   bit # is: |   30...24    |   23...17    |   16...10    |
   meaning:  | z bc_type id | y bc_type id | x bc_type id |

  ***************************************************************/

void FieldObj::init_boundary()
{
long      i;
int       p, bc = 0, boxid;
int       ix, iy, iz;

 for (i = 0; i < size; i++) ptype[i] =  _INSIDE_;   // default: point inside

 for (i = 0; i < size; i++)
   {
   split(i, ix, iy, iz);
   p = (*region).point_type(ix, iy, iz); 
   boxid = p >> 7;
   p = p & INSIDE_MASK;              // remove box id (keep bits 1 through 7)
   if (p == 0)                       // point is outside
     {
     ptype[i] = _OUTSIDE_; 
     elem[i] = 0.0;
     }
   else if (p & SURF_MASK)           // point on the surface
     {
     if (p & SURF_XMIN)
       if (bc_deriv[bc = get_bc(boxid,0)] != 0) ptype[i] |= (bc << XSHIFT) + SURF_XMIN;
       else  { ptype[i] = (bc << XSHIFT) + DIRICHLET; continue; }
     
     if (p & SURF_XMAX) 
       if (bc_deriv[bc = get_bc(boxid,1)] != 0) ptype[i] |= (bc << XSHIFT) + SURF_XMAX;
          else { ptype[i] = (bc << XSHIFT) + DIRICHLET; continue; } 

     if (p & SURF_YMIN) 
       if (bc_deriv[bc = get_bc(boxid,2)] != 0) ptype[i] |= (bc << YSHIFT) + SURF_YMIN;
          else { ptype[i] = (bc << XSHIFT) + DIRICHLET; continue; }
     
     if (p & SURF_YMAX) 
       if (bc_deriv[bc = get_bc(boxid,3)] != 0) ptype[i] |= (bc << YSHIFT) + SURF_YMAX;
          else { ptype[i] = (bc << XSHIFT) + DIRICHLET; continue; }

     if (p & SURF_ZMIN)
       if (bc_deriv[bc = get_bc(boxid,4)] != 0) ptype[i] |= (bc << ZSHIFT) + SURF_ZMIN;
       else { ptype[i] = (bc << XSHIFT) + DIRICHLET; continue; }
     
     if (p & SURF_ZMAX)
       if (bc_deriv[bc = get_bc(boxid,5)] != 0) ptype[i] |= (bc << ZSHIFT) + SURF_ZMAX;
       else { ptype[i] = (bc << XSHIFT) + DIRICHLET; continue; }

     } // end if (p & SURF_MASK)
  
  } // end for loop


 for (i = 0; i < size; i++)   // SECOND PASS: take care of Dirichlet bc
   if ( ptype[i] & DIRICHLET ) 
     {
     elem[i] = bc_coef[ bc = ptype[i] >> XSHIFT ];
     split(i, ix, iy, iz);
     p = region->point_type(ix, iy, iz); 
 
     if ( ix + 1 < xsize ) if ((region->point_type(ix+1,iy,iz) & _INSIDE_) && (ptype[i+1] & VARY_MASK))
       ptype[i + 1] |= (bc << XSHIFT) + SURF_XMIN; 

     if ( ix - 1 > 0 )     if ((region->point_type(ix-1,iy,iz) &  _INSIDE_) && (ptype[i-1] & VARY_MASK))
       ptype[i - 1] |= (bc << XSHIFT) + SURF_XMAX; 

     if ( iy + 1 < ysize ) if ((region->point_type(ix,iy+1,iz) & _INSIDE_) && (ptype[i+xsize] & VARY_MASK))
       ptype[i + xsize] |= (bc << YSHIFT) + SURF_YMIN; 
     
     if ( iy - 1 > 0 )   if ((region->point_type(ix,iy-1,iz) & _INSIDE_) && (ptype[i-xsize] & VARY_MASK))
       ptype[i - xsize] |= (bc << YSHIFT) + SURF_YMAX; 

     if ( iz + 1 < zsize ) if ((region->point_type(ix,iy,iz+1) & _INSIDE_) && (ptype[i+xysize] & VARY_MASK))
        ptype[i + xysize] |= (bc << ZSHIFT) + SURF_ZMIN; 
     
     if ( iz - 1 > 0 ) if ((region->point_type(ix,iy,iz-1) & _INSIDE_) && (ptype[i-xysize] & VARY_MASK))
        ptype[i - xysize] |= (bc << ZSHIFT) + SURF_ZMAX; 

  } // end for loop

}


//*****************************************************************************
/*
double FieldObj::interpolate(double x, double y, double z)
{
int ix = 0, iy = 0, iz = 0;

while (x > xcoord[ix] && ix < (xsize - 1) ) ix++;
while (y > ycoord[iy] && iy < (ysize - 1) ) iy++;
while (z > zcoord[iz] && iz < (zsize - 1) ) iz++;

long i = ix + xsize * iy + xysize * iz;

double p1 = (x - xcoord[ix-1]) / (xcoord[ix] - xcoord[ix-1]);
double p2 = (y - ycoord[iy-1]) / (ycoord[iy] - ycoord[iy-1]);
double p3 = (z - zcoord[iz-1]) / (zcoord[iz] - zcoord[iz-1]);

double res =      p1  *      p2  *      p3  * elem[i                     ] 
           + (1 - p1) * (1 - p2) * (1 - p3) * elem[i - 1 - xsize - xysize]
           +      p1  * (1 - p2) * (1 - p3) * elem[i     - xsize - xysize]
           + (1 - p1) *      p2  * (1 - p3) * elem[i - 1         - xysize]
           + (1 - p1) * (1 - p2) *      p3  * elem[i - 1 - xsize         ]
           +      p1  *      p2  * (1 - p3) * elem[i             - xysize]
           +      p1  * (1 - p2) *      p3  * elem[i     - xsize         ]
           + (1 - p1) *      p2  *      p3  * elem[i - 1                 ];

return res;
}
*/
//***********************************************************************************
/*
double FieldObj::get_value(double x, double y, double z)
{
return elem[location_to_index(x, y, z)];
}
*/
//***********************************************************************************

double FieldObj::sum()
{
double s = 0.0;
for(long i=0; i<size; i++) if (ptype[i] & REGION_MASK) s += elem[i];
return s;
}

//***********************************************************************************

double FieldObj::averageSpherical()
{
 double s = 0.0, v = 0.0, dv;

 for (int i = 1; i < xsize; i++ )

   if ( ! ((ptype[i] | ptype[i-1]) & _OUTSIDE_) )

     if ( xcoord[i-1] == 0.0 ) {
       v += ( dv = 2 / 3.0 * pow( xcoord[i], 3 ) );
       s += 3.0 / 4.0 * dv * (elem[i] + elem[i-1] / 3.0 );
     }
     else  {
       v += ( dv = xgrid[i-1] * (xcoord[i-1] * xcoord[i-1] + xcoord[i] * xcoord[i] ) );
       s += xgrid[i-1] * (elem[i-1] * xcoord[i-1] * xcoord[i-1] + elem[i] * xcoord[i] * xcoord[i]);
     }

return s / v ;
}

//***********************************************************************************
 
double FieldObj::averageConical()
{
 double s = 0.0, v = 0.0, Ith0, Ith1, Ir0, Ir1, rp, rm, thp, thm;

 for (int iy = 1; iy < ysize; iy++ )  {

   Ith0 = cos( ycoord[iy-1] ) - cos( ycoord[iy] );
   Ith1 = sin( ycoord[iy] ) - sin( ycoord[iy-1] ) + ycoord[iy-1] * Ith0 - ygrid[iy-1] * cos( ycoord[iy] );
   thp = Ith1 - ycoord[iy-1] * Ith0;
   thm = ycoord[iy] * Ith0 - Ith1;

   for (int ix = 1; ix < xsize; ix++ ) {
     long i = ix + xsize * iy;
     if ( ! ((ptype[i] | ptype[i-1] | ptype[i-xsize] | ptype[i-1-xsize] ) & _OUTSIDE_) ) {

       Ir0 = ( pow( xcoord[ix], 3) - pow( xcoord[ix-1], 3) ) / 3.0;
       Ir1 = ( pow( xcoord[ix], 4) - pow( xcoord[ix-1], 4) ) / 4.0;
       rp = Ir1 - xcoord[ix-1] * Ir0;
       rm = xcoord[ix] * Ir0 - Ir1;

       s += dxinv[ix-1] * dyinv[iy-1] * ( elem[i-1-xsize] * rm * thm + elem[i-1] * rm * thp + 
                                          elem[i-xsize] * rp * thm + elem[i] * rp * thp );
       v += dxinv[ix-1] * dyinv[iy-1] * ( rm + rp ) * ( thm + thp );
     }

   }
 }

return s / v ;
}

//***********************************************************************************
/* 
double FieldObj::averageConical()
{
 double s = 0.0, v = 0.0, dv, dvy;

 for (int iy = 1; iy < ysize; iy++ )  {

   dvy = cos(ycoord[iy-1]) - cos(ycoord[iy]);

   for (int ix = 1; ix < xsize; ix++ ) {
     long i = ix + xsize * iy;
     if ( ! ((ptype[i] | ptype[i-1] | ptype[i-xsize] | ptype[i-1-xsize] ) & _OUTSIDE_) ) {

       v += ( dv = dvy * ( pow(xcoord[ix],3) - pow(xcoord[ix-1],3) ) );
       s += dv * (elem[i-1-xsize]  + elem[i-1] + elem[i-xsize] + elem[i]);
     }
   }
 }

return 0.25 * s / v;
}
*/
//***********************************************************************************

double FieldObj::L2norm()
{
double s = 0.0;
for(long i=0; i<size; i++) if (ptype[i] & REGION_MASK) s += elem[i] * elem[i];
return s / double(size);
}

//***********************************************************************************

double FieldObj::average()
{
double s = 0.0, v = 0.0, dv, dvx, dvxy;
int    ix, iy, iz;
long   i = 0;
double *e;

for(ix = 1; ix < xsize; ix++)
  {
  dvx = xgrid[ix-1];
  for(iy = 1; iy < ysize; iy++) 
    {
    dvxy = dvx * ygrid[iy-1];
    for(iz = 1; iz < zsize; iz++) 
      {
      e = elem + (i = ix + xsize * iy + xysize * iz);
  
      if ( ! ((ptype[i] | ptype[i-1] | ptype[i-xsize] | ptype[i-xysize] | ptype[i-1-xsize] |
	       ptype[i-xsize-xysize] | ptype[i-1-xysize] | ptype[i-1-xsize-xysize]) & _OUTSIDE_) )
	{
	  v += ( dv = dvxy * zgrid[iz-1] );

          s += dv * (*e +
           *(e-1) + *(e-xsize) + *(e-xysize) +
           *(e-1-xsize) + *(e-xsize-xysize) + *(e-1-xysize) +
           *(e-1-xsize-xysize) );  
	}
      }
    }
  }
 return s / v * 0.125;
}

//**********************************************************************************
//***********************************************************************************
/*
void print3d(long *var)
{
for (int iz=0; iz < zsize; iz++, printf("\n"))
 for (int iy=0; iy < ysize; iy++, printf("\n"))
  for (int ix=0; ix < xsize; ix++) 
      printf("%ld ",var[ix + iy * xsize + iz * xysize]);
}
*/  
//***********************************************************************************

void FieldObj::print_ptype()
{
for (int iz=0; iz < zsize; iz++, printf("\n"))
 for (int iy=0; iy < ysize; iy++, printf("\n"))
  for (int ix=0; ix < xsize; ix++) 
      print_ptype(ix + iy * xsize + iz * xysize);
}

//***********************************************************************************
/*
void FieldObj::print3d(int n)
{
        char s1[20], s2[20];
        strcpy(s1,"%.");
        sprintf(s2,"%d",n);
        strcat(s1,s2);
        strcat(s1,"f ");

for (int iz=0; iz<zsize; iz++, printf("\n"))
 for (int iy=0; iy<ysize; iy++, printf("\n"))
  for (int ix=0; ix<xsize; ix++) printf(s1,elem[ix+iy*xsize+iz*xysize]);

}
*/
//***************************************************************************************
//                              X   D I R E C T I O N
//***************************************************************************************
/*
 Find vector Result_new (to be stored in array Result) satisfying

 (1 + c * Dx^2 - Lin) Result_new = ( cx * Dx^2 + cy * Dy^2 + cz * Dz^2 ) This + Result 

 | diag[1] sup[1]   0      0 :                    |                | right[1] |
 | sub[2]  diag[2] sup[2]  0 :            0       |                |          |
 |...........................:....................| x Result_new = |          |
 |                           : diag[n-1] sup[n-1] |                |          |
 |            0              : sub[n]    diag[n]  |                | right[n] |

 Operator Dx^2:   Dx^2 f[i] == dxminus[i] f[ix-1] + dxcenter[i] f[i] + dxplus[i] f[ix+1]
                          i == (ix,iy,iz)

 sub[i] = c dxminus[i+i0]
 diag[i] = 1 + c dxcenter[i+i0] - Lin[i+i0]
 sup[i] = c dxplus[i+i0]
 right[i] = Result[i] + This[i] ( cx dxcenter[i] + cy dycenter[i] + cz dzcenter[i]) +
                     cx This[ix-1] dxminus[i] + cy This[iy-1] dyminus[i] + cz This[iz-1] dzminus[i] +
                     cx This[ix+1] dxplus[i]  + cy This[iy+1] dyplus[i]  + cz This[iz+1] dzplus[i]
                      
 Special case of Dirichlet boundary ( df/dx[i] + bc_lin f[i] = bc_coef ):
 Assume uniform for accuracy!!

  |       |      |    f[1] - f[-1] = 2 xgrid[0] ( bc_coef - f[0] bc_lin )
virtual f[0]   f[1]   Eliminating f[-1]: f[-1] = f[1] - 2 xgrid[0] ( bc_coef - f[0] bc_lin )
 node

  Dx^2 f[0] = ( -2 f[0] + f[1] + f[-1] ) / xgrid[0]^2
            = 2 ( f[1] - f[0] ) / xgrid[0]^2 - 2 ( bc_coef - f[0] bc_lin ) / xgrid[0]
            = 2 f[0] ( bc_lin - 1 / xgrid[0] ) / xgrid[0] +          | -> diag[0]
              2 f[1] / xgrid[0]^2 -                                  | -> sup[0]
              2 bc_coef / xgrid[0]                                   | -> right[0]

 c Dx^2 new[0] - cx Dx^2 old[0] = new[0] diag[0] + new[1] sup[1] - right[0]

 sup[0] = 2 c / xgrid[0]^2
 diag[0] = 1 - Lin + 2 c ( bc_lin - 1 / xgrid[0] ) / xgrid[0] = 1 - Lin + 2 c bc_lin / xgrid[0] - sup[0]

 right[0] = 2 c bc_coef / xgrid[0] + 
            2 cx / xgrid[0] ( (old[1] - old[0]) / xgrid[0] + old[0] bc_lin - bc_coef )
          = 2 / xgrid[0]^2 ( cx ( old[1] + old[0] ( xgrid[0] bc_lin - 1 ) ) + xgrid[0] bc_coef (c - cx) )

 */
//***************************************************************************************

void FieldObj::xRun(double c, double cx, double cy, double cz, double *Lin, double *Result)
{
int      ix = 0, iy = 0, iz = 0;
long     i = 0, istart = 0, bcvar;
int      n = 0, bc;
double   dg, sp, sb, ri;
double   cx2 = 2 * cx, cy2 = 2 * cy, cz2 = 2 * cz, c2 = 2 * c;

if ( D <= 0.0 )
  {
  for (i = 0; i < size; i ++)
    if (ptype[i] & VARY_MASK) 
      Result[i] /= (1.0 -  (Lin ? Lin[i] : 0.0) + 1.0e-16);
  return;
  }

do
  {
  while ( (bcvar = ptype[i]) & VARY_MASK )
    {
    n++;
    dg = 1 + c * dxcenter[ix] - (Lin ? Lin[i] : 0.0);
    sp = c * dxplus[ix]; sb = c * dxminus[ix];
    ri = Result[i] + elem[i] * ( cx * dxcenter[ix] + cy * dycenter[iy] + cz * dzcenter[iz] );

    //******************* x direction check **************************

    if ( bcvar & SURF_XMIN )
      {
      istart = i; 
      if ( bc_deriv[bc = (bcvar >> XSHIFT) & BC_ID_MASK ] == 0)  
        ri += bc_coef[bc] * (cx - c) * dxminus[ix] + cx * elem[i+1] * dxplus[ix];
      else 
        {
	ri += dxinv[ix] * (c2 * bc_coef[bc] + 
   	      cx2 * ( (elem[i+1] - elem[i]) * dxinv[ix] + elem[i] * bc_lin[bc] - bc_coef[bc] ) ) - 
              cx * dxcenter[ix] * elem[i];
        dg += c * ( 2 * bc_lin[bc] * dxinv[ix] - dxcenter[ix] ) - ( sp = c2 * dx2inv[ix] );
	}
      }
    else if ( bcvar & SURF_XMAX ) 
      if ( bc_deriv[bc = (bcvar >> XSHIFT) & BC_ID_MASK ] == 0)  
          ri += bc_coef[bc] * (cx - c) * dxplus[ix] + cx * elem[i-1] * dxminus[ix];
      else 
        {
	ri += dxinv[ix-1] * (c2 * bc_coef[bc] + 
	    cx2 * ( (elem[i-1] - elem[i]) * dxinv[ix-1] + elem[i] * bc_lin[bc] - bc_coef[bc] ) ) - 
            cx * dxcenter[ix] * elem[i];
        dg += c * ( 2 * bc_lin[bc] * dxinv[ix-1] - dxcenter[ix] ) - ( sb = c2 * dx2inv[ix-1] );
        }

    else ri += cx * (dxminus[ix] * elem[i-1] + dxplus[ix] * elem[i+1]);  //***  not on X surface

    //******************** y direction check **************************

    if ( bcvar & SURF_YMIN )       
      if ( bc_deriv[bc = (bcvar >> YSHIFT) & BC_ID_MASK ] == 0) 
        ri += cy * ( dyminus[iy] * bc_coef[bc] + dyplus[iy] * elem[i+xsize] );
      else ri += cy2 * ( (elem[i+xsize] - elem[i]) * dy2inv[iy] + dyinv[iy] * ( elem[i] * bc_lin[bc] - bc_coef[bc] ) ) -
                 cy * dycenter[iy] * elem[i];

    else 
      if ( bcvar & SURF_YMAX ) //* bcvar = SURF_YMAX    
        if ( bc_deriv[bc = (bcvar >> YSHIFT) & BC_ID_MASK ] == 0) 
          ri += cy * ( dyplus[iy] * bc_coef[bc] + dyminus[iy] * elem[i-xsize]);
        else ri += cy2 * ( (elem[i-xsize] - elem[i]) * dy2inv[iy-1] + dyinv[iy-1] * ( elem[i] * bc_lin[bc] - bc_coef[bc] ) ) -
             cy * dycenter[iy] * elem[i];

    else ri += cy * (dyminus[iy] * elem[i-xsize] + dyplus[iy] * elem[i+xsize]);   //***  not on Y surface

    //********************* z direction check **************************

    if ( bcvar & SURF_ZMIN )
       if ( bc_deriv[bc = (bcvar >> ZSHIFT) & BC_ID_MASK] == 0) 
         ri += cz * ( dzminus[iz] * bc_coef[bc] + dzplus[iz] * elem[i+xysize] );
       else ri += cz2 * ( (elem[i+xysize] - elem[i]) * dz2inv[iz] + dzinv[iz] * ( elem[i] * bc_lin[bc] - bc_coef[bc] ) ) -
                  cz * dzcenter[iz] * elem[i];

    else  //* bcvar = SURF_ZMAX
      if ( bcvar & SURF_ZMAX )
        if ( bc_deriv[bc = (bcvar >> ZSHIFT) & BC_ID_MASK] == 0)
          ri += cz * ( dzplus[iz] * bc_coef[bc] + dzminus[iz] * elem[i-xysize] );
        else ri += cz2 * ( (elem[i-xysize] - elem[i]) * dz2inv[iz-1] + dzinv[iz-1] * ( elem[i] * bc_lin[bc] - bc_coef[bc] ) ) -
                   cz * dzcenter[iz] * elem[i];

    else ri += cz * (dzminus[iz] * elem[i-xysize] + dzplus[iz] * elem[i+xysize]);  //***  not on Z surface
         
    //******************************************************************

    right[n] = ri; diag[n] = dg; sup[n] = sp; sub[n] = sb;
    i ++;
    if (++ix == xsize) break;

    } // while bcvar > 0 and ix < xsize
 
  if (n)
    {
    tridiag(n);
    for (int ii = 1, ind = istart; ii<=n; ii++, ind ++) Result[ind] = right[ii];
    n = 0;
    }

  if (ix < xsize) 
    while ( !(ptype[i] & VARY_MASK) )  
      {
      if ( ptype[i] & DIRICHLET ) Result[i] = bc_coef[ptype[i] >> XSHIFT];
      i ++; 
      if ( (++ix) == xsize ) break; 
      }

  if (ix == xsize)       
     { 
     ix = 0;
     if (++iy == ysize) { iy = 0; iz ++; } 
     }
  
  } while (iz < zsize);   

}

//***************************************************************************************
//                               Y   D I R E C T I O N
//***************************************************************************************
//
//  (1 + c * Dy^2 - Lin) Result_new = ( cx * Dx^2 + cy * Dy^2 + cz * Dz^2 ) This + Result
//
//***************************************************************************************

void FieldObj::yRun(double c, double cx, double cy, double cz, double *Lin, double *Result)
{
int      ix = 0, iz = 0, iy = 0;
long     i = 0, istart = 0, bcvar;
int      n = 0, bc;
double   dg, sp, sb, ri;
double   cx2 = 2 * cx, cy2 = 2 * cy, cz2 = 2 * cz, c2 = 2 * c;

if ( D <= 0.0 )
  {
  for (i = 0; i < size; i ++) if (ptype[i] & VARY_MASK) 
     Result[i] /= (1.0 -  (Lin ? Lin[i] : 0.0) + 1.0e-12);
  return;
  }

do
  {
  while ( (bcvar = ptype[i]) & VARY_MASK )
    {
    n++;
    dg = 1 + c * dycenter[iy] - (Lin ? Lin[i] : 0);
    sp = c * dyplus[iy];
    sb = c * dyminus[iy];
    ri = Result[i] + elem[i] * ( cx * dxcenter[ix] + cy * dycenter[iy] + cz * dzcenter[iz] );

    //******************* y direction check **************************

    if ( bcvar & SURF_YMIN )
      {
      istart = i; 
      if ( bc_deriv[bc = (bcvar >> YSHIFT) & BC_ID_MASK] == 0)
        ri += bc_coef[bc] * dyminus[iy] * (cy - c) + cy * dyplus[iy] * elem[i+xsize];
      else 
        {
	ri += dyinv[iy] * (c2 * bc_coef[bc] + 
   	    cy2 * ( (elem[i+xsize] - elem[i]) * dyinv[iy] + elem[i] * bc_lin[bc] - bc_coef[bc] ) ) - 
            cy * dycenter[iy] * elem[i];
        dg += c * ( 2 * bc_lin[bc] * dyinv[iy] - dycenter[iy] ) - ( sp = c2 * dy2inv[iy] );
	}
      }
    else if ( bcvar & SURF_YMAX ) //* bcvar = SURF_YMAX
   
      if ( bc_deriv[bc = (bcvar >> YSHIFT) & BC_ID_MASK] == 0)
        ri += bc_coef[bc] * dyplus[iy] * (cy - c) + cy * dyminus[iy] * elem[i-xsize];
      else 
        {
        ri += dyinv[iy-1] * (c2 * bc_coef[bc] + 
	     cy2 * ( (elem[i-xsize] - elem[i]) * dyinv[iy-1] + elem[i] * bc_lin[bc] - bc_coef[bc] ) ) - 
             cy * dycenter[iy] * elem[i];
        dg += c * ( 2 * bc_lin[bc] * dyinv[iy-1] - dycenter[iy] ) - ( sb = c2 * dy2inv[iy-1] );
        }

    else ri += cy * (dyminus[iy] * elem[i-xsize] + dyplus[iy] * elem[i+xsize]);  // not on Y surface

    //******************** x direction check **************************

    if ( bcvar & SURF_XMIN )       
      if ( bc_deriv[bc = (bcvar >> XSHIFT) & BC_ID_MASK] == 0) 
        ri += cx * ( dxminus[ix] * bc_coef[bc] + dxplus[ix] * elem[i+1] );
      else ri += cx2 * ( (elem[i+1] - elem[i]) * dx2inv[ix] + dxinv[ix] * ( elem[i] * bc_lin[bc] - bc_coef[bc] ) ) -
                 cx * dxcenter[ix] * elem[i];
	      
    else  if ( bcvar & SURF_XMAX ) 
      if ( bc_deriv[bc = (bcvar >> XSHIFT) & BC_ID_MASK] == 0) 
        ri += cx * ( dxplus[ix] * bc_coef[bc] + dxminus[ix] * elem[i-1] );
      else ri += cx2 * ( (elem[i-1] - elem[i]) * dx2inv[ix-1] + dxinv[ix-1] * ( elem[i] * bc_lin[bc] - bc_coef[bc] ) ) -
                 cx * dxcenter[ix] * elem[i];

    else ri += cx * (dxminus[ix] * elem[i-1] + dxplus[ix] * elem[i+1]);    // not on X surface

    //********************* z direction check **************************

    if ( bcvar & SURF_ZMIN )
      if ( bc_deriv[bc = (bcvar >> ZSHIFT) & BC_ID_MASK] == 0) 
        ri += cz * ( dzminus[iz] * bc_coef[bc] + dzplus[iz] * elem[i+xysize] );
      else ri += cz2 * ( (elem[i+xysize] - elem[i]) * dz2inv[iz] + dzinv[iz] * ( elem[i] * bc_lin[bc] - bc_coef[bc] ) ) -
                 cz * dzcenter[iz] * elem[i];

    else  if ( bcvar & SURF_ZMAX ) //* bcvar = SURF_ZMAX
      if ( bc_deriv[bc = (bcvar >> ZSHIFT) & BC_ID_MASK] == 0)
        ri += cz * ( dzplus[iz] * bc_coef[bc] + dzminus[iz] * elem[i-xysize] );
      else ri += cz2 * ( (elem[i-xysize] - elem[i]) * dz2inv[iz-1] + dzinv[iz-1] * ( elem[i] * bc_lin[bc] - bc_coef[bc] ) ) -
                 cz * dzcenter[iz] * elem[i];

    else ri += cz * (dzminus[iz] * elem[i-xysize] + dzplus[iz] * elem[i+xysize]);  //***  not on Z surface
         
    //******************************************************************

    right[n] = ri; diag[n] = dg;
    sup[n] = sp;   sub[n] = sb;
    i += xsize;  
    if (++iy == ysize) break;

    } // while bcvar > 0 and iy < ysize

    if (n > 0)
      {
      tridiag(n);
      for (int ii = 1, ind = istart; ii<=n; ii++, ind += xsize) Result[ind] = right[ii]; 
      n = 0;
      }

    if (iy < ysize)
      while ( !(ptype[i] & VARY_MASK) )  
         {  
         if ( ptype[i] & DIRICHLET ) Result[i] = bc_coef[ptype[i] >> XSHIFT];
         i += xsize; if ( (++iy) == ysize ) break; 
         }

    if (iy == ysize) 
      {
      iy = 0 ;
      if (++iz == zsize) { i -= size - 1; iz = 0; ix++; }
      }

  } while (ix < xsize); 
}


//***************************************************************************************
//                              Z   D I R E C T I O N
//***************************************************************************************
//
// (1 + c * Dz^2 - Lin) Result_new = ( cx * Dx^2 + cy * Dy^2 + cz * Dz^2 ) This + Result 
//
//***************************************************************************************

void FieldObj::zRun(double c, double cx, double cy, double cz, double *Lin, double *Result)

{
int      iz = 0, ix = 0, iy = 0;
long     i = 0, istart = 0, bcvar;
int      n = 0, bc;
double   dg, sp, sb, ri;
double   cx2 = 2 * cx, cy2 = 2 * cy, cz2 = 2 * cz, c2 = 2 * c;

if ( D <= 0.0 )
  {
  for (i = 0; i < size; i ++) if (ptype[i] & VARY_MASK) 
    Result[i] /= (1.0 - (Lin ? Lin[i] : 0.0) + 1.0e-12);
  return;
  }

do
  {
  while ( (bcvar = ptype[i]) & VARY_MASK )
    {  
    n++;
    dg = 1 + dzcenter[iz] * c - (Lin ? Lin[i] : 0.0);
    sp = c * dzplus[iz];
    sb = c * dzminus[iz];
    ri = Result[i] + elem[i] * ( cx * dxcenter[ix] + cy * dycenter[iy] + cz * dzcenter[iz] );

    //******************* z direction check **************************
          
    if ( bcvar & SURF_ZMIN )
      {
      istart = i;
      if ( bc_deriv[bc = (bcvar >> ZSHIFT) & BC_ID_MASK] == 0)  
        ri += dzminus[iz] * bc_coef[bc] * (cz - c) + dzplus[iz] * cz *  elem[i+xysize];
      else 
        {
        ri += dzinv[iz] * (c2 * bc_coef[bc] + 
   	    cz2 * ( (elem[i+xysize] - elem[i]) * dzinv[iz] + elem[i] * bc_lin[bc] - bc_coef[bc] ) ) -
            cz * dzcenter[iz] * elem[i];
        dg += c * ( 2 * bc_lin[bc] * dzinv[iz] - dzcenter[iz] ) - ( sp = c2 * dz2inv[iz] );
	}
      }
      else  if ( bcvar & SURF_ZMAX )            
        if ( bc_deriv[bc = (bcvar >> ZSHIFT) & BC_ID_MASK] == 0)
           ri += dzplus[iz] * bc_coef[bc] * (cz - c) + cz * dzminus[iz] * elem[i-xysize];
        else 
          {
	  ri += dzinv[iz-1] * (c2 * bc_coef[bc] + 
   	      cz2 * ( (elem[i-xysize] - elem[i]) * dzinv[iz-1] + elem[i] * bc_lin[bc] - bc_coef[bc] ) ) - 
              cz * dzcenter[iz] * elem[i];
          dg += c * ( 2 * bc_lin[bc] * dzinv[iz-1] - dzcenter[iz] ) - ( sb = c2 * dz2inv[iz-1] );
	  }
       else ri += cz * (dzminus[iz] * elem[i-xysize] + dzplus[iz] * elem[i+xysize]); //** not on Z surface

    //******************** x direction check **************************

    if ( bcvar & SURF_XMIN )       
      if ( bc_deriv[bc = (bcvar >> XSHIFT) & BC_ID_MASK] == 0) 
        ri += cx * ( dxminus[ix] * bc_coef[bc] + dxplus[ix] * elem[i+1] );
      else ri += cx2 * ( (elem[i+1] - elem[i]) * dx2inv[ix] + dxinv[ix] * ( elem[i] * bc_lin[bc] - bc_coef[bc] ) ) -
                 cx * dxcenter[ix] * elem[i];
	      
    else  if ( bcvar & SURF_XMAX ) 
      if ( bc_deriv[bc = (bcvar >> XSHIFT) & BC_ID_MASK] == 0) 
        ri += cx * ( dxplus[ix] * bc_coef[bc] + dxminus[ix] * elem[i-1] );
      else ri += cx2 * ( (elem[i-1] - elem[i]) * dx2inv[ix-1] + dxinv[ix-1] * ( elem[i] * bc_lin[bc] - bc_coef[bc] ) ) -
                 cx * dxcenter[ix] * elem[i];

    else ri += cx * (dxminus[ix] * elem[i-1] + dxplus[ix] * elem[i+1]);    // not on X surface

      //********************* y direction check **************************

    if ( bcvar & SURF_YMIN )       
      if ( bc_deriv[bc = (bcvar >> YSHIFT) & BC_ID_MASK ] == 0) 
        ri += cy * ( dyminus[iy] * bc_coef[bc] + dyplus[iy] * elem[i+xsize] );
      else ri += cy2 * ( (elem[i+xsize] - elem[i]) * dy2inv[iy] + dyinv[iy] * ( elem[i] * bc_lin[bc] - bc_coef[bc] ) ) -
                 cy * dycenter[iy] * elem[i];

    else 
      if ( bcvar & SURF_YMAX ) //* bcvar = SURF_YMAX    
        if ( bc_deriv[bc = (bcvar >> YSHIFT) & BC_ID_MASK ] == 0) 
          ri += cy * ( dyplus[iy] * bc_coef[bc] + dyminus[iy] * elem[i-xsize]);
        else ri += cy2 * ( (elem[i-xsize] - elem[i]) * dy2inv[iy-1] + dyinv[iy-1] * ( elem[i] * bc_lin[bc] - bc_coef[bc] ) ) -
             cy * dycenter[iy] * elem[i];

    else ri += cy * (dyminus[iy] * elem[i-xsize] + dyplus[iy] * elem[i+xsize]);   //***  not on Y surface

    //******************************************************************

    right[n] = ri; diag[n] = dg;
    sup[n] = sp;   sub[n] = sb;
    i += xysize; ++iz;
    if (iz == zsize) break;

    } // while ptype > 0 and iz < zsize

  if (n > 0) 
    {  
    tridiag(n);
    for (int ii = 1, ind = istart; ii<=n; ii++, ind += xysize) Result[ind] = right[ii]; 
    n = 0;
    } 
 
  if (iz < zsize)
    while ( !(ptype[i] & VARY_MASK) )  
       {
       if ( ptype[i] & DIRICHLET ) Result[i] = bc_coef[ptype[i] >> XSHIFT];
       i += xysize; if ( (++iz) == zsize) break; 
       }

  if (iz == zsize) 
      {
      iz = 0; i -= (size - 1);
      if (++ix == xsize) { ix = 0; iy ++; }
      }

  } while (iy < ysize);

}

/***************************************************************************************
                          1 D   S P H E R I C A L   R U N
***************************************************************************************

 Find vector Result_new (to be stored in array Result) satisfying

 (1 + c * Dx^2 - Lin) Result_new = cc * Dx^2 This + Result 

 | diag[1] sup[1]   0      0 :                    |                | right[1] |
 | sub[2]  diag[2] sup[2]  0 :            0       |                |          |
 |...........................:....................| x Result_new = |          |
 |                           : diag[n-1] sup[n-1] |                |          |
 |            0              : sub[n]    diag[n]  |                | right[n] |

 Operator Dx^2:   Dx^2 f[i] == minus f[i-1] + center f[i] + plus f[i+1] + coef
 off the boundary: minus = dxminus[i], center = dxcenter[i], plus = dxplus[i]

 sub[i]   = c minus
 diag[i]  = 1 + c center - Lin[i]
 sup[i]   = c plus
 right[i] = Result[i] + cc ( elem[i] center + elem[i-1] minus + elem[i+1] plus ) + (cc - c) coef
                                       
 Dx^2 f = 1/r^2 d( r^2 df/dr) / dr = 

      1   r[i+1/2]^2 (f[i+1]-f[i]) / dr[i] - r[i-1/2]^2 (f[i] - f[i-1]) / dr[i-1] 
   ------ ----------------------------------------------------------------------- =
   r[i]^2                          ( dr[i] + dr[i-1] ) / 2

           2                       r[i+1/2]^2          r[i-1/2]^2          r[i+1/2]^2   r[i-1/2]^2
 ------------------------ { f[i+1] ---------- + f[i-1] ---------- - f[i] ( ---------- + ---------- ) }
 r[i]^2 (dr[i] + dr[i-1])             dr[i]              dr[i-1]              dr[i]       dr[i-1]

 Special case of Dirichlet boundary ( df/dx[i] + bc_lin f[i] = bc_coef ):
 Assume uniform for accuracy!!

  |       |      |    f[1] - f[-1] = 2 xgrid[0] ( bc_coef - f[0] bc_lin )
virtual f[0]   f[1]   Eliminating f[-1]: f[-1] = f[1] - 2 xgrid[0] ( bc_coef - f[0] bc_lin )
 node

               1           (r + dr/2)^2         (r - dr/2)^2          (r + dr/2)^2   (r - dr/2)^2
 Dx^2 f[0] = ------ { f[1] ------------ + f[-1] ------------ - f[0] ( ------------ + ------------ ) }
             r^2 dr             dr                  dr                     dr             dr

      1                           4 r^2               2 r
  = ----- { ( f[1] - f[0] ) ( 1 + ------ ) - dr ( 1 - ---- )^2 ( bc_coef - bc_lin * f[0] ) }
    2 r^2                          dr^2                dr


 */
//***************************************************************************************

void FieldObj::spheRun( double c, double cc, double *Lin, double *Result)
{
long     i = 0, istart = 0, bcvar;
int      n = 0;
double   dg, sp, sb, ri;
double   plus, minus, center, coef;

if ( D <= 0.0 )
  {
  for (i = 0; i < size; i ++)
    if (ptype[i] & VARY_MASK) Result[i] /= (1.0 -  (Lin ? Lin[i] : 0.0) + 1.0e-16);
  return;
  }

do
  {
  while ( (bcvar = ptype[i]) & VARY_MASK )
    {
    n++;

    dg = 1 - (Lin ? Lin[i] : 0.0);
    ri = Result[i];
    minus = dxminus[i]; plus = dxplus[i]; center = dxcenter[i];
    coef = 0.0;

    //******************* x direction check **************************

    nablaR(i, bcvar, minus, center, plus, coef);

    if ( bcvar & SURF_XMIN )   istart = i; 

    dg += c * center; 
    ri += (cc - c) * coef + cc * center * elem[i];
    if ( ! (bcvar & SURF_XMIN) ) { sb = c * minus; ri += cc * minus * elem[i-1]; }
    if ( ! (bcvar & SURF_XMAX) ) { sp = c * plus; ri += cc * plus * elem[i+1]; }    

    right[n] = ri; diag[n] = dg; sup[n] = sp; sub[n] = sb;
    if (++i == xsize) break;

    } // while bcvar > 0 and i < xsize
 
  if (n)
    {
    tridiag(n);
    for (int ii = 1, ind = istart; ii<=n; ii++, ind ++) Result[ind] = right[ii];
    n = 0;
    }

  if (i < xsize)
    while ( !(ptype[i] & VARY_MASK) )  
       {
       if ( ptype[i] & DIRICHLET ) Result[i] = bc_coef[ptype[i] >> XSHIFT];
       if ( ++i == xsize) break; 
       }

  } while (i < xsize);

  }
/***************************************************************************************
      CONICAL GEOMETRY RUN, THETA DIRECTION:
                                                           1 
 Solving  (1 + c Dr^2 - Lin) Result_new = ( cr Dr^2 + cth --- Dth^2 ) This + Result
                                                          r^2
                 1   d        df
 Dr^2 f[i,j] == --- --- ( r^2 -- ) == dxminus[i] f[i-1] + dxcenter[i] f[i] + dxplus[i] f[i+1]
                r^2  dr       dr 
 where r == r[i]

                      1       d              df
 Dth^2 f[i,j] == ----------- ---- ( sin(th) ---- ) == dyminus[j] f[j-1] + dycenter[j] f[j] + dyplus[j] f[j+1]
                 r^2 sin(th) d th           d th
 where th == theta[j]
 

 At Neumann boundary:  f[j-1] = f[j] - 2 dygrid[j] (bc_coef - bc_lin f[j])

                    1                           2 sin(theta) cos(dth/2)   
 Dth^2 f[j] = --------------- ( (f[j+1] - f[j]) ----------------------- - 2 sin(th - dth/2) (bc_coef - bc_lin f[j]) )
              r^2 sin(th) dth                            dth

  


 ***************************************************************************************/

    //******************* calculating D_r^2 *************************

void FieldObj::nablaR(int ix, long bcvar, 
                      double &rminus, double &rcenter, double &rplus, double &rcoef)
{
  double a, b;
  int bc;

    rminus = dxminus[ix]; rplus = dxplus[ix]; rcenter = dxcenter[ix];
    rcoef = 0.0;

    if ( bcvar & SURF_XMIN )
      {
      rminus = 0.0;
      if ( bc_deriv[bc = (bcvar >> XSHIFT) & BC_ID_MASK ] == 0)  
        rcoef = dxminus[ix] * bc_coef[bc];
      else 
        if (xcoord[ix] > 0 ) { 
          a = 2 * xcoord[ix] * dxinv[ix];
          b = 0.5 / ( xcoord[ix] * xcoord[ix] );
          rplus = b * (1 + a * a );
          rcoef = - b * (1 - a) * (1 - a) * xgrid[ix]; 
          rcenter = - (rplus + rcoef * bc_lin[bc]);
          rcoef *= bc_coef[bc];
 	  }
        else {
          rcoef = - 6 * dxinv[ix] * bc_coef[bc];
          rcenter = - ( rplus = 6 * dx2inv[ix] ) + 6 * dxinv[ix] * bc_lin[bc];
	  }
      }
    else if ( bcvar & SURF_XMAX ) 
      {
      rplus = 0.0;
      if ( bc_deriv[bc = (bcvar >> XSHIFT) & BC_ID_MASK ] == 0)  
         rcoef = dxplus[ix] * bc_coef[bc];
      else 
        {
        a = 2 * xcoord[ix] * dxinv[ix-1];
        b = 0.5 / ( xcoord[ix] * xcoord[ix] );
        rminus = b * ( 1 + a * a );
        rcoef = - b * (1 - a) * (1 - a) * xgrid[ix-1]; 
        rcenter = - (rminus + rcoef * bc_lin[bc]);
        rcoef *= bc_coef[bc];
        }
      }

}
    //******************* calculating D_th^2 *************************

void FieldObj::nablaTheta(int iy, long bcvar, 
                      double &thminus, double &thcenter, double &thplus, double &thcoef)
{
  int bc;

    thminus = dyminus[iy]; thplus = dyplus[iy]; thcenter = dycenter[iy];
    thcoef = 0.0;

    if (bcvar & SURF_YMIN) {
        thminus = 0.0;
        if ( bc_deriv[bc = (bcvar >> YSHIFT) & BC_ID_MASK ] == 0)  
          thcoef = dyminus[iy] * bc_coef[bc];
        else 
          if (ycoord[iy] != 0) { 
            thplus = 2 * cos( 0.5 * ygrid[iy] ) * dy2inv[iy];
            thcoef = - 2 * dyinv[iy] * sin( ycoord[iy] - 0.5 * ygrid[iy]) / sin( ycoord[iy] ) ;
            thcenter = - (thplus + thcoef * bc_lin[bc]);
            thcoef *= bc_coef[bc];
 	    }
	  else  {  // theta = 0 
            thplus = 4 * dy2inv[iy];
            thcoef = - 4 * dyinv[iy];
            thcenter = - (thplus + thcoef * bc_lin[bc]);
            thcoef *= bc_coef[bc];
 	    }
         }
     else  if (bcvar & SURF_YMAX) {
        thplus = 0.0;
        if ( bc_deriv[bc = (bcvar >> YSHIFT) & BC_ID_MASK ] == 0)  
          thcoef = dyplus[iy] * bc_coef[bc];
        else 
          if ( fabs( ycoord[iy] - pi) > 1.0e-12 ) { 
            thminus = 2 * cos( 0.5 * ygrid[iy-1] ) * dy2inv[iy-1];
            thcoef = - 2 * dyinv[iy-1] * sin( ycoord[iy] - 0.5 * ygrid[iy-1]) / sin( ycoord[iy] ) ;
            thcenter = - (thminus + thcoef * bc_lin[bc]);
            thcoef *= bc_coef[bc];
 	    }
	  else  {  // theta =  pi 
            thminus = 4 * dy2inv[iy-1];
            thcoef = - 4 * dyinv[iy-1];
            thcenter = - (thminus + thcoef * bc_lin[bc]);
            thcoef *= bc_coef[bc];
 	    }
     }

}
//***************************************************************************************

void FieldObj::rRun( double c, double cr, double cth, double *Lin, double *Result)
{
long     i = 0, istart = 0, bcvar;
int      ix = 0, iy = 0, iyold = -1;
int      n = 0;
double   dg, sp, sb, ri;
double   rplus, rminus, rcenter, rcoef;
double   thplus, thminus, thcenter, thcoef;
double   thp, thm, thcr, thcf;
double   a;
static   const double pi = atan(1) * 4;

if ( D <= 0.0 )
  {
  for (i = 0; i < size; i ++)
    if (ptype[i] & VARY_MASK) Result[i] /= (1.0 - (Lin ? Lin[i] : 0.0) + 1.0e-16);
  return;
  }
/*
 if (xcoord[0] < 1e-20) {
   double s=0.0, v=0.0, dv, ave;
   for (iy = 0; iy < ysize; iy++) {
     v += ( dv = cos(ycoord[iy] - ( iy ? 0.5 * ygrid[iy-1] : 0.0 )) - 
                 cos(ycoord[iy] + ( iy < ysize - 1 ? 0.5 * ygrid[iy] : 0.0 )) );
     s += dv * Result[iy * xsize];
   }
   ave = s / v;
   for (iy = 0; iy < ysize; iy++) Result[iy * xsize] = ave;
 }
 iy = 0;
*/
do
  {
  while ( (bcvar = ptype[i]) & VARY_MASK )
    {
    n++;

    dg = 1 - (Lin ? Lin[i] : 0.0);
    ri = Result[i];

    if ( bcvar & SURF_XMIN )   istart = i; 

    nablaR(ix, bcvar, rminus, rcenter, rplus, rcoef);     // calculating D_r^2 
    if (iy != iyold) {
      iyold = iy;
      nablaTheta(iy, bcvar, thm, thcr, thp, thcf);    // calculating D_theta^2
    }

    if (xcoord[ix] == 0) a = 0.0;   // no nabla theta at origin
    else a = 1 / ( xcoord[ix] * xcoord[ix] );

    //*******************  build matrix  ******************************

    thplus = a * thp; thminus = a * thm; thcenter = a * thcr; thcoef = a * thcf;
    dg += c * rcenter; 
    ri += (cr - c) * rcoef + cth * thcoef + (cr * rcenter + cth * thcenter) * elem[i];
    if ( ! (bcvar & SURF_XMIN) ) { sb = c * rminus; ri += cr * rminus * elem[i-1]; }
    if ( ! (bcvar & SURF_XMAX) ) { sp = c * rplus;  ri += cr * rplus * elem[i+1]; }    
    if ( ! (bcvar & SURF_YMIN) )  ri += cth * thminus * elem[i-xsize]; 
    if ( ! (bcvar & SURF_YMAX) )  ri += cth * thplus * elem[i+xsize];    

    //    fprintf(stderr,"(%d,%d): Dx=(%g,%g,%g,%g), Dy=(%g,%g,%g,%g) a=%.4g\n",ix,iy,rminus,rcenter,rplus,rcoef,
    //	    thminus,thcenter,thplus,thcoef,a);

    right[n] = ri; diag[n] = dg; sup[n] = sp; sub[n] = sb;
    ++i;
    if (++ix == xsize) break;

    } // while bcvar > 0 and i < xsize
 
  if (n)
    {
    tridiag(n);
    for (int ii = 1, ind = istart; ii<=n; ii++, ind ++) Result[ind] = right[ii];
    n = 0;
    }

  if (ix < xsize)
    while ( !(ptype[i] & VARY_MASK) )  
       {
       if ( ptype[i] & DIRICHLET ) Result[i] = bc_coef[ptype[i] >> XSHIFT];
       ++i;
       if ( ++ix == xsize) break; 
       }

  if (ix == xsize) {
    ix = 0;
    iy ++;
  }

  } while (iy < ysize);

}

//**********************************************************************************************

void FieldObj::thetaRun( double c, double cr, double cth, double *Lin, double *Result)
{
long     i = 0, istart = 0, bcvar;
int      ix = 0, iy = 0, ixold = -1;
int      n = 0;
double   dg, sp, sb, ri;
double   rplus, rminus, rcenter, rcoef;
double   thplus, thminus, thcenter, thcoef;
double   a;
static   const double pi = atan(1) * 4;

if ( D <= 0.0 )
  {
  for (i = 0; i < size; i ++)
    if (ptype[i] & VARY_MASK) Result[i] /= (1.0 - (Lin ? Lin[i] : 0.0) + 1.0e-16);
  return;
  }

do
  {
  while ( (bcvar = ptype[i]) & VARY_MASK )
    {
    n++;

    dg = 1 - (Lin ? Lin[i] : 0.0);
    ri = Result[i];

    if ( bcvar & SURF_YMIN )   istart = i; 

    nablaTheta(iy, bcvar, thminus, thcenter, thplus, thcoef);   // calculating D_theta^2
    if (ix != ixold) {
      ixold = ix;
      nablaR(ix, bcvar, rminus, rcenter, rplus, rcoef);     // calculating D_r^2  
      if (xcoord[ix] == 0) a = 0.0;   // no nabla theta at origin
      else a = 1 / ( xcoord[ix] * xcoord[ix] );
   }
   
    //*******************  build matrix  ******************************

    thplus *= a; thminus *= a; thcenter *= a; thcoef *= a;
    dg += c * thcenter; 
    ri += cr * rcoef + (cth - c) * thcoef + (cr * rcenter + cth * thcenter) * elem[i];
    if ( ! (bcvar & SURF_XMIN) ) ri += cr * rminus * elem[i-1];
    if ( ! (bcvar & SURF_XMAX) ) ri += cr * rplus * elem[i+1];   
    if ( ! (bcvar & SURF_YMIN) ) { sb = c * thminus; ri += cth * thminus * elem[i-xsize];  }
    if ( ! (bcvar & SURF_YMAX) ) { sp = c * thplus;  ri += cth * thplus * elem[i+xsize];   }     

    right[n] = ri; diag[n] = dg; sup[n] = sp; sub[n] = sb;

    i += xsize;
    if (++iy == ysize) break;

    } // while bcvar > 0 and iy < ysize
 
  if (n)
    {
    tridiag(n);
    for (int ii = 1, ind = istart; ii<=n; ii++, ind += xsize) Result[ind] = right[ii];
    n = 0;
    }

  if (iy < ysize)
    while ( !(ptype[i] & VARY_MASK) )  
       {
       if ( ptype[i] & DIRICHLET ) Result[i] = bc_coef[ptype[i] >> XSHIFT];
       i += xsize;
       if ( ++iy == ysize) break; 
       }

  if (iy == ysize) {
    iy = 0;
    i = ++ix;
  }

  } while (ix < xsize);

}
     
//**********************************************************************************************
/*

 void FieldObj::dirichlet()
    { 
    int ix, iy, iz;
    for (long i=0; i<size; i++)  
      if (ptype[i] & DIRICHLET) elem[i] = bc_coef[ptype[i] >> XSHIFT];
    }

*/
//***************************************************************

void FieldObj::print_ptype(long i)
{
long x = ptype[i];
if (x & _OUTSIDE_) printf("[OUT] ");
else
  {
  if (x & VARY_MASK) printf("[VARY");
  else if (x & DIRICHLET) printf("[f=%g", bc_coef[x >> XSHIFT]); 
  if (x & SURF_MASK)
    {
    if (x & SURF_XMIN) printf(" %s XMIN", bc_id[(x >> XSHIFT) & BC_ID_MASK]);
    if (x & SURF_XMAX) printf(" %s XMAX", bc_id[(x >> XSHIFT) & BC_ID_MASK]);
    if (x & SURF_YMIN) printf(" %s YMIN", bc_id[(x >> YSHIFT) & BC_ID_MASK]);
    if (x & SURF_YMAX) printf(" %s YMAX", bc_id[(x >> YSHIFT) & BC_ID_MASK]);
    if (x & SURF_ZMIN) printf(" %s ZMIN", bc_id[(x >> ZSHIFT) & BC_ID_MASK]);
    if (x & SURF_ZMAX) printf(" %s ZMAX", bc_id[(x >> ZSHIFT) & BC_ID_MASK]);
    }
   printf("] ");
  }
}

//***************************************************************************

void FieldObj::tridiag(int n)
{
int    i;
register double sb, sp, dg, rt;

sp = ( sup[1] /= (dg = diag[1]) );
rt = ( right[1] /= dg );

for (i=2; i<=n; i++)
  {
  sb = sub[i];
  rt = right[i] = (right[i] - sb * rt) / (dg = diag[i] - sb * sp);
  sp = (sup[i] /= dg);
  }

for (i=n-1; i>=1; i--) right[i] -= sup[i] * right[i+1];
}

//****************************************************************************
//*                   C L A S S   B U F F E R   A R R A Y
//****************************************************************************

  BufferObj :: BufferObj(TokenString &params, const char *id) : 
  FieldObj(params, id)
    {
    if (total <= 0) fprintf(stderr,">> BufferObj::BufferObj: WARNING: [%s] = %g\n", ID,total);

    KD = kplus = kminus = -1.0;

    if (params.token2_count(id, "kminus")) 
      params.trail2_pars(id, "kminus", 1, 'd', &kminus, 'E'); 
    if (params.token2_count(id, "kplus")) 
      params.trail2_pars(id, "kplus",  1, 'd', &kplus,  'E'); 
    if (params.token2_count(id, "KD")) 
      params.trail2_pars(id, "KD",  1, 'd', &KD,  'E'); 

    if (kplus >= 0 && kminus >= 0) KD = kminus / kplus;
    else if (kplus >= 0 && KD >= 0) kminus = kplus * KD;
    else if (kminus >= 0 && KD >= 0)  kplus  = kminus / KD;
    else {
      fprintf(stderr,"\n*** ERROR: incomplete kinetics specifications for buffer \"%s\"\n",ID);
      exit(1);
    }

    double calcium = params.trail2_double("Ca","total");
    bgr = total / (1.0 + kplus / kminus * calcium);

    if (VERBOSE) 
      {
      fprintf(stderr,"    Kinetics for buffer %s: KD=%g uM, k-=%g /ms, k+=%g /(ms uM)\n",
	                 ID, KD, kminus, kplus);
      fprintf(stderr,"    Free background [%s] = %g uM\n", ID, bgr);
      }

    if (!params.token2_count(id, "import")) *this = bgr;

    for (int j = 0; j < region->get_surface_num(); j++)
        if (bc_deriv[bccond[j]] == 0)  bc_coef[bccond[j]] = bgr;
     }

//****************************************************************************

BufferArray::BufferArray(TokenString &params, const char *group)
    {
    ID = StrCpy( group);
    buf_num = params.token_count(group);
    array = new BufferObj *[buf_num];

    if ( !buf_num )  return;

    for (int i=0; i<buf_num; i++)
      {
      char id[40];
      params.trail_string(ID, id, i+1);
      array[i] = new BufferObj(params, id);
      }
    };

//****************************************************************************

void BufferArray::allocateCopy(const BufferArray &bufs)
    {
    if (bufs.ID) ID = StrCpy( bufs.ID );
    if ( (buf_num = bufs.buf_num) > 0 )
      {
      array = new BufferObj *[buf_num];
      for (int i = 0; i < buf_num; i++) array[i] = new BufferObj(*bufs.array[i]);
      }
    }

//****************************************************************************
