#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "Ygl.h"
#include <string.h>
#include "vector.h"
#include "syntax.h"
#include "box.h"
#include "field.h"
#include "expr.h"
#include "table.h"
#include "peak.h"
#include "interpol.h"
#include "gate.h"
#include "fplot.h"

extern  double *xgrid, *ygrid, *zgrid;
extern  double *xcoord, *ycoord, *zcoord;

extern  int   xsize, ysize, zsize;
extern  long  xysize;
extern  int    VERBOSE;

extern  long   location_to_index(double, double, double);
extern  double nearest_location(double *, double *, int, double);

extern  void split(long, int &, int &, int &);
extern  double *ResolveID(TokenString &Param, int p, double **,
                         FieldObj *Ca, BufferArray *Bufs, class KineticObj *KO);
void mouse_wait() 
    { gflush(); 
    while (!getbutton(RIGHTMOUSE)); while (getbutton(RIGHTMOUSE)); 
    };

void mouse_pause() 
    { gflush(); while (getbutton(LEFTMOUSE)); }

//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
//        C L A S S   P L O T O B J E C T   I M P L E M E N T A T I O N
//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@


void PlotObj::draw_grid(double x0, double x1, double y0, double y1, char lines)
{
  char xformat[20];
  char yformat[20];

  int xlines = int( double(plot_width) / double(pixelspergridline) + 0.5);
  int ylines = int( double(plot_height) / double(pixelspergridline) + 0.5);

  double dx = format_double(x1 - x0, xlines, 1, xformat);
  double dy = format_double(y1 - y0, ylines, 1, yformat);

  double yscale = double(plot_height) / double(y1 - y0);
  double xscale = double(plot_width) / double(x1 - x0);

  winset(winid);
  color(grid_color);
  sboxs(xmargin - 1, ymargin - 1, xmargin + plot_width + 1, ymargin + plot_height + 1);
  color(text_color);  

  double x, y;

  x = dx * floor(x0 / dx);  if (x0 > 0) x += dx;
  y = dy * floor(y0 / dy);  if (y0 > 0) y += dy;

  //printf("x=%g x1-x0=%g dx=%g y=%g y0=%g y1=%g y1-y0=%g dy=%g \n", x, x1-x0, dx, y, y0, y1, y1-y0, dy);

  char s[25];
  int xx, yy;

  while ( x <= x1 )
    { 
    xx = int(xscale * (x - x0) + 0.5);
    if (lines) { 
               color(grid_color); 
               move2s(xx + xmargin, ymargin);
               draw2s(xx + xmargin, ymargin + plot_height); 
               }
    sprintf(s,xformat,x);
    cmov2s(xmargin + xx - 4, plot_height + ymargin + 4);
    color(text_color); charstr(s);
    x += dx;
    }

 while ( y <= y1 )
   { 
   yy = int(yscale * (y - y0) + 0.5); 
   if (lines) { 
              color(grid_color);
              move2s(xmargin, yy + ymargin);
              draw2s(xmargin + plot_width, ymargin + yy); 
              }
   sprintf(s,yformat,y);
   cmov2s(xmargin + plot_width + 3, ymargin + yy - 8);
   color(text_color); charstr(s);
   y += dy;
   }
}


//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
//*******************************************************************************
//            C L A S S   2 D - P L O T   I M P L E M E N T A T I O N
//*******************************************************************************

Plot2D::Plot2D(FieldObj *f, char islog, char dir, double coord, double total, int steps, int hor, int ver, 
               double min, double max, const char *label) : FieldPlotObj(f, islog, "", label)
{
tscale = double(steps) / total;

if (max > min)
  {
  min_fixed = min; max_fixed = max;
  fixed = 1;
  df_dcol_fixed = (max_fixed - min_fixed) / double(color_num);
  }
else fixed = 0;

switch (dir)
  {
  case 'x': grid1 = ygrid;   grid2 = zgrid;
            coord1 = ycoord; coord2 = zcoord;
            incr1 = xsize;   incr2 = xysize; 
            num1  = ysize;   num2  = zsize;
            coord = nearest_location(xcoord, xgrid, xsize, coord);
            field_index = location_to_index(coord, coord1[0], coord2[0]);
            break;
 
  case 'y': grid1 = xgrid;   grid2 = zgrid;
            coord1 = xcoord; coord2 = zcoord;
            incr1 = 1;       incr2 = xysize; 
            num1  = xsize;   num2  = zsize;
            coord = nearest_location(ycoord, ygrid, ysize, coord);
            field_index = location_to_index(coord1[0], coord, coord2[0]);
            break; 

  case 'z': grid1 = xgrid;   grid2 = ygrid;
            coord1 = xcoord; coord2 = ycoord;
            incr1 = 1;       incr2 = xsize; 
            num1 = xsize;    num2  = ysize;
            coord = nearest_location(zcoord, zgrid, zsize, coord);
            field_index = location_to_index(coord1[0], coord2[0], coord);
            break;
  }
    sprintf(win_title,"%s on %c-plane at %.3g", f->ID, dir, coord);

    double f1 = double(hor) / (coord1[num1-1] - coord1[0]);
    double f2 = double(ver) / (coord2[num2-1] - coord2[0]);
    double factor = ( (f1 > f2) ? f2 : f1);

    plot_width  = int(factor * (coord1[num1-1] - coord1[0]) + 0.5);
    plot_height = int(factor * (coord2[num2-1] - coord2[0]) + 0.5);

    win_width  = plot_width  + 2 * xmargin + char_width * 6 + 4;
    win_height = plot_height + 2 * ymargin + dy_color_panel + 3 * char_height + 10; 
    
    prefsize(win_width, win_height);
    winid = winopen(win_title);
    cmode();         
    gconfig();  qdevice(ESCKEY);
    putenv("YGL_PCM=1");

    set_colors();
    redraw();
}

//*************************************************************************************

void Plot2D::redraw()
{
  winset(winid);
  color(bgr_color);
  clear();
  color(plot_color);
  sboxfs(xmargin, ymargin, xmargin + plot_width, ymargin + plot_height);
  draw_grid(coord1[0], coord1[num1-1], coord2[0], coord2[num2-1], 0);
  x_value = field->Time - 0.000001;
  //mouse_pause();
  draw();
}

//*************************************************************************************

void Plot2D::set_colors()
{
  mapcolor(grid_color,  200, 150, 150);
  mapcolor(bgr_color,   236, 226, 236);
  mapcolor(plot_color,  210, 230, 220);
  mapcolor(text_color,  0,   0,   0);
  mapcolor(label_color, 255, 0,   0);

  int red, green, blue;
  color1 = 255 - color_num;

  for(int i=color1; i<= color1 + color_num; i++)  
      {
      red = green = int( 255.0 * (i - color1) / double(color_num) );
      blue = 255 - red;
      mapcolor(i, red, green, blue);
      }
}

//*************************************************************************************

void Plot2D::get_range()
{
long   ind;
double f;

  fmin = get_value(field_index);
  fmax = get_value(field_index);


  for (int j = 0; j < num2; j++)
    {
    for (int i = 0; i < num1; i++)  
      {
      ind = field_index + j * incr2 + i * incr1;
      if (field->ptype[ind] & REGION_MASK)
        {
	f = get_value(ind);
        if (f < fmin) fmin = f; else if (f > fmax) fmax = f;
	}
      }
    }
  if ( (fabs( (fmax-fmin)/(fmax+fmin+1.0e-12)) < 1e-16 ) || (fmax <= fmin) )
     fmax = (fmin + 1e-8)  * (1 + 1e-8);
  df_dcol = (fmax - fmin) / double(color_num);
}

//*************************************************************************************

void Plot2D::color_panel()
{
int  y0, y1, x0, x1;
char mintxt[40], smin[40];
char maxtxt[40], smax[40];
char form[40];
 
  y0 = ymargin + plot_height + getheight() + 10;
  y1 = y0 + dy_color_panel;
  x0 = xmargin - 4;

  color(bgr_color);
  sboxfs(x0 - 1, y0 - 1, win_width, y1 + 1);

  get_range();

  format_double(fmax-fmin, 1, 3, form);  
  sprintf(smin, form, fmin);
  sprintf(smax, form, fmax);

  sprintf(mintxt, "min=%s", smin);
  sprintf(maxtxt, "max=%s", smax);

  color(text_color);
  cmov2(x0, y0), charstr(mintxt);
 
  x0 = x0 + strwidth(mintxt) + 4;  // x0 coordinate for color panel
  x1 = win_width - (xmargin - 4) - strwidth(maxtxt);  // coordinate for "fmax="

  cmov2(x1, y0), charstr(maxtxt);

  x1 -= 6;  // x1 coordinate for color panel

  double dfdx = (fmax - fmin) / double(x1 - x0);
  double f;

  color(grid_color);
  sboxfs(x0-1, y0-1, x1 + 1, y1 + 1);

  for (int xn = x0; xn <= x1; xn++ )
    {
    f = fmin + dfdx * (xn - x0);
    color(get_color(f)); 
    pnt2s(xn, y0); draw2s(xn,y1);
    }

}

//*************************************************************************************
//                          P L O T  2 D   D R A W
//*************************************************************************************

void Plot2D::draw()
{
long   ind;
char   text1[80], text2[80];
static char form[80];

  double t = field->Time;
  //  fprintf(stderr, "t=%g x_value=%g tscale=%g \n",t,x_value,tscale);
  if ( int(t * tscale) == int(x_value * tscale) ) return; 
 
  winset(winid);
  color_panel();

  int ychar = ymargin + plot_height + getheight() + dy_color_panel + 15; 
  int xchar = xmargin - 4;

  color(bgr_color);
  sboxfs(xchar - 1, ychar - 1, win_width, win_height);

  format_double(t - x_value, 1, 1, form);
  x_value = t;
  sprintf(text1, form, x_value);
  sprintf(text2, "%s%s", x_label, text1);
  cmov2s(xchar, ychar);  color(label_color);
  charstr(text2);

  double scale1 = double(plot_width) / (coord1[num1-1] - coord1[0] + 0.5 * ( grid1[0] + grid1[num1-1]) );
  double scale2 = double(plot_height) / (coord2[num2-1] - coord2[0] + 0.5 * ( grid2[0] + grid2[num2-1]) );

  int c;
  int    x, y, xx, yy;
  double dx, dy;
  y = ymargin;

  for (int j = 0; j < num2; j++)
    {
    x = xmargin;
    dy = coord2[j] - coord2[0] + 0.5 * ( grid2[0] + grid2[ j ? j - 1 : 0] ); 
    yy = int( dy * scale2 + 0.5) + ymargin;
    for (int i = 0; i < num1; i++)  
      {
      ind = field_index + j * incr2 + i * incr1;
      dx = coord1[i] - coord1[0] + 0.5 * ( grid1[0] + grid1[ i ? i - 1 : 0] ); 
      xx = int(dx * scale1 + 0.5) + xmargin;
      if (field->ptype[ind] & REGION_MASK)
        {
        c = get_color( get_value(ind) );
        if (c > 0) { color( c ); sboxfs(x, y, xx, yy); }
        }
      x = xx;
      }
    y = yy;
    } 
  mouse_pause();
}

//*************************************************************************************

int Plot2D::get_color(double f)
{
int c;
if (! fixed)
  {
  c = color1 + int( (f - fmin) / df_dcol + 0.5 );
  if (c < color1 || c > 255) 
    printf("Plot2D::get_color: color out of range: f=%g fmin=%g fmax=%g\n", f, fmin, fmax);
  }
else
  {
  c = color1 + int( (f - min_fixed) / df_dcol_fixed + 0.5 );
  if (c > 255) c = color_num; 
    else if (c < color1) c = color1; 
  }
return c;
}

//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
//*******************************************************************************
//       B A S E   F I E L D   P L O T   1 D   I M P L E M E N T A T I O N
//*******************************************************************************

FieldPlot1D::FieldPlot1D(FieldObj *f, char islog, char dir, double coord1, double coord2,
               const char *label) :  FieldPlotObj(f, islog, "", label)
{
switch (dir)
  {
  case 'x': grid = xgrid; coord = xcoord; incr = 1; num = xsize;
            coord1 = nearest_location(ycoord, ygrid, ysize, coord1);
            coord2 = nearest_location(zcoord, zgrid, zsize, coord2);
            field_index = location_to_index(coord[0], coord1, coord2);
            sprintf(win_title,"[%s] along x-axis (y=%.3g, z=%.3g)", f->ID, coord1, coord2);
            break;
 
  case 'y': grid = ygrid; coord = ycoord; incr = xsize; num = ysize;
            coord1 = nearest_location(xcoord, xgrid, xsize, coord1);
            coord2 = nearest_location(zcoord, zgrid, zsize, coord2);
            field_index = location_to_index(coord1, coord[0], coord2);
            sprintf(win_title,"[%s] along y-axis (x=%.3g, z=%.3g)", f->ID, coord1, coord2);
            break;

  case 'z': grid = zgrid; coord = zcoord; incr = xysize; num = zsize;
            coord1 = nearest_location(xcoord, xgrid, xsize, coord1);
            coord2 = nearest_location(ycoord, ygrid, ysize, coord2);

            field_index = location_to_index(coord1, coord2, coord[0]);
            sprintf(win_title,"[%s] along z-axis (x=%.3g, y=%.3g)", f->ID, coord1, coord2);
            break;
  }
}

//*************************************************************************************

void FieldPlot1D::get_range()
{
long   ind;
double f;

  for (int j = 0; j < num; j++)  {
    ind = field_index + j * incr;
    if (field->ptype[ind] & REGION_MASK)  {
      f = get_value(ind);
      if (f < fmin) fmin = f;  
        else if (f > fmax) fmax = f;
    }
  }
    
  if ( (fabs((fmax - fmin) / (fmax+fmin+1.0e-12)) < 1.0e-16) || (fmax <= fmin) )
     fmax = (fmin + 1e-8) * (1 + 1e-8);
}

//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
//*******************************************************************************
//                   P L O T   1 D   I M P L E M E N T A T I O N
//*******************************************************************************

Plot1D::Plot1D(FieldObj *f, char islog, char dir, double coord1, double coord2,
               double total, int n, int hor, int ver, const char *label) :  
        FieldPlot1D(f, islog, dir, coord1, coord2, label)
{
  tscale = double(n) / total;

  plot_width = hor;
  plot_height = ver;

  win_width = plot_width + 2 * xmargin + char_width * 6 + 3;
  win_height = plot_height + 2 * ymargin + char_height * 2 + 6;

  prefsize(win_width, win_height);
  winid = winopen(win_title);
  cmode();  gconfig(); qdevice(ESCKEY);
  putenv("YGL_PCM=1");

  set_colors();
  redraw();
}


//*************************************************************************************

void Plot1D::set_colors()
{
  mapcolor(grid_color,  255,255,249);
  mapcolor(bgr_color,   236,230,236);
  mapcolor(plot_color,  210,205,210);
  mapcolor(text_color,  0, 0, 0);
  mapcolor(label_color, 0, 0, 255);
  mapcolor(line_color,  255, 0, 255);
}

//*************************************************************************************

void Plot1D::get_range()
{
  fmin = get_value(field_index);
  fmax = get_value(field_index); 

  FieldPlot1D::get_range();

  if ( fabs(fmax - fmin) < 1.0e-15 ) fmax = fmin + 1.0e-14;
}

//*************************************************************************************
//                           P L O T   1 D   D R A W
//*************************************************************************************

void Plot1D::draw()
{
long   ind;
int    x, y;
char   text[80], xtxt[80], form[80];
char   smin[40], smax[40], mintxt[40], maxtxt[40];

  double t = field->Time;
  if ( int(tscale * x_value) == int(tscale * t) ) return;

  winset(winid);
  color(bgr_color);  clear();
  color(plot_color);
  sboxfs(xmargin, ymargin, xmargin + plot_width, ymargin + plot_height);
  get_range();
  draw_grid(coord[0], coord[num-1], fmin, fmax, 1);

  int ychar = ymargin + plot_height + getheight() + 7; 
  int xchar = xmargin - 4;

  color(bgr_color);
  sboxfs(xchar - 1, ychar - 1, win_width, win_height);

  format_double(fmax-fmin, 1, 4, form);  
  sprintf(smin, form, fmin);
  sprintf(smax, form, fmax);
  sprintf(mintxt, "min=%s", smin);
  sprintf(maxtxt, "max=%s", smax);

  format_double(t - x_value, 1, 1, form);
  x_value = t;
  sprintf(xtxt, form, x_value);
  sprintf(text, "%s%s %s %s", x_label, xtxt, mintxt, maxtxt);
  cmov2s(xchar, ychar);  color(label_color);
  charstr(text);

  color(line_color);
  linewidth(2);
  x = xmargin;
  ind = field_index;
  pnt2s(xmargin, 
        ymargin + int( plot_height * (get_value(ind) - fmin) / (fmax - fmin) + 0.5 ));

  double scale = double(plot_width) / (coord[num-1] - coord[0]);

  for (int j = 0; j < num; j++, ind += incr)
    {
    x = xmargin + int( (coord[j] - coord[0]) * scale + 0.5);
    y = ymargin;
    if (field->ptype[ind] & REGION_MASK) 
      y += int( plot_height * (get_value(ind) - fmin) / (fmax - fmin) + 0.5 );
    draw2s(x,y);
    }

  linewidth(1);
  mouse_pause();
}
//*************************************************************************************

void Plot1D::print()
{
  int ind = field_index;

  char fname[80];

  sprintf(fname,"%s%ld.1Dplot.xmgr", field->ID, field_index);
  FILE *file = fopen(fname,"w");

  sprintf(fname,"%s%ld.1Dplot.dat", field->ID, field_index);
  FILE *fdat = fopen(fname,"w");

  fprintf(file,"@ with g0 \n@ g0 on \n");
  fprintf(file,"@ title \"%s\"\n", win_title);
  fprintf(file,"@ title font 7\n");
  fprintf(file,"@ xaxis label \"distance (\\8m\\4m) \"\n");
  fprintf(file,"@ yaxis label \"[%s] (\\8m\\4M) \"\n", field->ID);

  fprintf(file,"@ target s0 \n@ type xy\n");

  for (int j = 0; j < num; j++, ind += incr)
    {
    fprintf(file, "%g %g\n", coord[j], get_value(ind));
    fprintf(fdat, "%g %g\n", coord[j], get_value(ind));
    }

  fprintf(file,"&\n@ autoscale \n@ redraw\n");
  fclose(file);
}

//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
//*******************************************************************************
//        C L A S S   R A S T E R    P L O T   I M P L E M E N T A T I O N
//*******************************************************************************

RasterPlot::RasterPlot(FieldObj *f, char islog, char dir, double coord1, double coord2, double t,
           int n, int hor, int ver,  double min, double max, 
           const char *label) : FieldPlot1D(f, islog, dir, coord1, coord2, label)
{
if (max > min)
  {
  min_fixed = min; max_fixed = max;
  fixed = 1;
  df_dcol_fixed = (max_fixed - min_fixed) / double(color_num);
  }
else fixed = 0;

  fmin = get_value(field_index);
  fmax = get_value(field_index); 

  steps = n;
  count = 0; 
  raster = new double[(steps + 2) * num + 1];
  times  = new double[steps + 2];
  
  for (int j = 0; j < num; j++) raster[j] = get_value(field_index + j * incr);
  times[0] = f -> Time; 
  last = times[0] + t;
 
  plot_width = hor; plot_height = ver;

  win_width = plot_width + 2 * xmargin + char_width * 6 + 4;
  win_height = plot_height + 2 * ymargin + dy_color_panel + 3 * char_height + 10;

  prefsize(win_width, win_height);
  winid = winopen(win_title);
  cmode(); gconfig(); qdevice(ESCKEY);
  putenv("YGL_PCM=1");

  set_colors();
  redraw();
}

//*************************************************************************************

void RasterPlot::redraw()
{
  winset(winid);
  color(bgr_color);
  clear();
  color(plot_color);
  sboxfs(xmargin, ymargin, xmargin + plot_width, ymargin + plot_height);
  draw_grid(times[0], last, coord[0], coord[num-1], 0);
  if (count) count --;
  draw();
}

//*************************************************************************************

void RasterPlot::set_colors()
{
  mapcolor(grid_color,  200, 150, 150);
  mapcolor(bgr_color,   236,226,236);
  mapcolor(plot_color,  210, 205, 210);
  mapcolor(text_color,  0, 0, 0);
  mapcolor(label_color, 255, 0, 0);

  int red, green, blue;
  color1 = 255 - color_num;

  for(int i=color1; i<= color1 + color_num; i++)  
      {
      red = green = int( 255.0 * (i - color1) / double(color_num) );
      blue = 255 - red;
      mapcolor(i, red, green, blue);
      }
}

//*************************************************************************************

void RasterPlot::get_range()
{
  FieldPlot1D::get_range();
  df_dcol = (fmax - fmin) / double(color_num);
}

//*************************************************************************************

void RasterPlot::color_panel()
{
int  y0, y1, x0, x1;
char mintxt[40], smin[40];
char maxtxt[40], smax[40];
char form[40];
 
  y0 = ymargin + plot_height + getheight() + 10;
  y1 = y0 + dy_color_panel;
  x0 = xmargin - 4;

  color(bgr_color);
  sboxfs(x0 - 1, y0 - 1, win_width, y1 + 1);

  format_double(fmax-fmin, 1, 3, form);  
  sprintf(smin, form, fmin);
  sprintf(smax, form, fmax);

  sprintf(mintxt, "min=%s", smin);
  sprintf(maxtxt, "max=%s", smax);

  color(text_color);
  cmov2(x0, y0), charstr(mintxt);
 
  x0 = x0 + strwidth(mintxt) + 4;  // x0 coordinate for color panel
  x1 = win_width - (xmargin - 4) - strwidth(maxtxt);  // coordinate for "fmax="

  cmov2(x1, y0), charstr(maxtxt);

  x1 -= 6;  // x1 coordinate for color panel

  double dfdx = (fmax - fmin) / double(x1 - x0);
  double f;

  color(grid_color);
  sboxfs(x0-1, y0-1, x1 + 1, y1 + 1);

  for (int xn = x0; xn <= x1; xn++ )
    {
    f = fmin + dfdx * (xn - x0);
    color(get_color(f)); 
    pnt2s(xn, y0); draw2s(xn,y1);
    }

}

//*************************************************************************************
//                         R A S T E R   P L O T   D R A W
//*************************************************************************************

void RasterPlot::draw()
{
char   text1[80], text2[80], form[80];

  double t = field->Time;

  if (count > 0) 
    {
    while (t < times[count]) count--;
    if ( int( times[count] * double(plot_width) / last )
       ==  int( t * double(plot_width) / last ) ) return;
    }

  if (t > times[count] && count < steps )  times[++count] = t;
  else if (count >= steps) printf("RasterPlot::draw: buffer overflow [steps=%d]\n",steps);

  for (int i = 0; i < num; i++)
      raster[ count * num + i ] = get_value(field_index + i * incr);

  get_range();  
  winset(winid);
  color_panel();

  int ychar = ymargin + plot_height + getheight() + dy_color_panel + 15; 
  int xchar = xmargin - 4;

  color(bgr_color);
  sboxfs(xchar - 1, ychar - 1, win_width, win_height);

  format_double(times[count] - times[count-1], 1, 1, form);
  sprintf(text1, form, times[count]);
  sprintf(text2, "%s%s", x_label, text1);
  cmov2s(xchar, ychar);  color(label_color);
  charstr(text2);

  double scale1 = double(plot_width) / (last - times[0]);
  double scale2 = double(plot_height) / (coord[num-1] - coord[0]);

  int    x, y, xx, yy;
  double dy;

  y = ymargin;

  if (count == 0) return;
 
  for (int j = 0; j < num; j++)
    {
    x = xmargin;
    dy = ( j + 1 < num ? coord[j+1] - 0.5*grid[j] : coord[num-1] ) - coord[0]; 
    yy = int( dy * scale2 + 0.5) + ymargin;
    for (int i = 0; i < count; i++)
      {  
      xx = xmargin + int( (0.5 * (times[i] + times[i+1]) - times[0]) * scale1 + 0.5);
      if (xx > x) if (field->ptype[field_index + j * incr] & REGION_MASK)
        { 
        color( get_color( raster[i * num + j] ) );
        sboxfs(x, y, xx, yy);
        }
      x = xx;
      }
    y = yy;
    }
  mouse_pause();
}

//*************************************************************************************

int RasterPlot::get_color(double f)
{
int c;
if (! fixed)
  {
  c = color1 + int( (f - fmin) / df_dcol + 0.5 );
  if (c < color1 || c > 255) {  }
  if (c > 255) c = color_num; 
    else if (c < color1) c = color1; 
  }
else
  {
  c = color1 + int( (f - min_fixed) / df_dcol_fixed + 0.5 );
  if (c > 255) c = color_num; 
    else if (c < color1) c = color1; 
  }
return c;
}

//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
//*******************************************************************************
//         C L A S S   P O I N T   P L O T   I M P L E M E N T A T I O N
//*******************************************************************************

PointPlot::PointPlot(double *ptr, double *tptr, char islog, double t, int n, 
  const char *txt, const char *label, int hor, int ver, const char *fn) 
  : PlotObj(ptr, islog, txt, label)
{
  fname = StrCpy(fn);
  timeptr = tptr;

  fmin = fmax = get_value(0);
  fmax += 1.0e-12;

  count = 0;
  steps = n;
 
  buffer = new double[steps + 2];  buffer[0] = get_value(0);
  times  = new double[steps + 2]; 
  times[0] = *tptr;
  last = t + times[0];

  plot_width = hor;
  plot_height = ver;

  win_width = plot_width + 2 * xmargin + char_width * 6 + 3;
  win_height = plot_height + 2 * ymargin + char_height * 2 + 6;

  prefsize(win_width, win_height);
  winid = winopen(win_title);
  cmode();         
  gconfig(); qdevice(ESCKEY);
  putenv("YGL_PCM=1");

  set_colors();
  redraw();
}


//*************************************************************************************

void PointPlot::set_colors()
{
  mapcolor(grid_color,  255,255,249);
  mapcolor(bgr_color,   236,230,236);
  mapcolor(plot_color,  210,205,210);
  mapcolor(text_color,  0, 0, 0);
  mapcolor(label_color, 0, 0, 255);
  mapcolor(line_color,  255, 0, 255);
}

//*************************************************************************************
/*
void PointPlot::get_range()
{
  if ( get_value(0) < fmin )  fmin = get_value(0);
  if ( get_value(0) > fmax )  fmax = get_value(0);   
    
  if ( (fabs((fmax - fmin)/(fmax + fmin + 1.0e-12)) < 1.0e-16) || (fmax <= fmin) ) 
    fmax = (fmin + 1e-8) * (1 + 1e-8);
}
*/
//*************************************************************************************
//                         T I M E   P L O T   D R A W
//*************************************************************************************

void PointPlot::draw()
{
int    x, xn, y, yn;
char   txt[200], xform[80], xtxt[80], format[80];
char   mmform[30], mintxt[30], maxtxt[40];
static double tscale = double(plot_width) / last;

  double t = *timeptr;

  while (t < times[count] && count) count--;
  if  ( count && (int( t * tscale ) == int( times[count] * tscale ) ) ) {
    if ( get_value(0) < fmin )  fmin = buffer[count] = get_value(0);
    if ( get_value(0) > fmax )  fmax = buffer[count] = get_value(0);   
    return;
  }
  if ( get_value(0) < fmin )  fmin = get_value(0);
  if ( get_value(0) > fmax )  fmax = get_value(0);   

  if (count < steps && t > times[count])  ++count;
  else if (count >= steps) {                           // if buffer overflow
    double *newtimes = new double[steps + 400 + 2];
    double *newbuffer = new double[steps + 400 + 2];
    steps += 400;
    for (int i = 0; i <= steps; i++) { newtimes[i] = times[i]; newbuffer[i] = buffer[i]; }
    delete [] times; delete [] buffer;
    times = newtimes; buffer = newbuffer;
    count ++;
  } 

  times[count] = t;
  buffer[count] = get_value(0);

  winset(winid);
  color(bgr_color);  clear();
  color(plot_color);
  sboxfs(xmargin, ymargin, xmargin + plot_width, ymargin + plot_height);

  //printf("%g %g %e %e\n",times[0], last, fmin, fmax);
  draw_grid(times[0], last, fmin, fmax, 1);

  int ychar = ymargin + plot_height + getheight() + 7; 
  int xchar = xmargin - 4;

  color(bgr_color);
  sboxfs(xchar - 1, ychar - 1, win_width, win_height);

  strcpy(xform, x_label);
  format_double(times[count] - times[count-1], 1, 1, format);
  strcat(xform,format);

  format_double(fmax-fmin, 1, 4, mmform);
  sprintf(mintxt, mmform, fmin);
  sprintf(maxtxt, mmform, fmax);
  sprintf(xtxt, xform, times[count]);

  sprintf(txt, "%s min=%s max=%s", xtxt, mintxt, maxtxt);
  cmov2s(xchar, ychar);  color(label_color);
  charstr(txt);

  color(line_color);
  linewidth(2);
  double xscale = double(plot_width) / (last - times[0]);
  double yscale = double(plot_height) / (fmax - fmin);

  pnt2s(xmargin, ymargin + int( yscale * (buffer[0] - fmin) + 0.5 ) );
  yn = x = 0;

  for (int j = 1; j <= count; j++)
    {
    xn = xmargin + int( xscale * (times[j] - times[0]) + 0.5 );
    y  = ymargin + int( yscale * (buffer[j] - fmin) + 0.5 );
    if (y > yn) yn = y;
    if (xn > x) { draw2s(xn, yn); yn = 0; }
    x = xn;
    }

  linewidth(1);
  mouse_pause();
}

//*************************************************************************************

void PointPlot::print()
{
  FILE *f;

  if ( fname[0] ) 
   { if ( !(f = fopen(fname,"w")) ) printf("PointPlot::print: cannot open %s\n",fname); }
  else 
   { if ( !(f = fopen(win_title,"w")) ) printf("PointPlot::print: cannot open %s\n",win_title); }

  // fprintf(f,"# %s\n",win_title);

  for (int j = 0; j <= count; j++)
      fprintf(f,"%g %g\n", times[j], buffer[j]);

  fclose(f);

}

//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
//*******************************************************************************
//             B A S E   C L A S S   X M G R   P I P E   P L O T
//*******************************************************************************

 int XmgrPlot::graph_num = 0;
 int XmgrPlot::graph_count = 0;
 int XmgrPlot::rows = 0;
 int XmgrPlot::cols = 0;
 FILE *XmgrPlot::file = stdout;

//*******************************************************************************

  void XmgrPlot::init(int gnum, FILE *f, int rs, int cs)
    {
    if ( !(graph_num = gnum) ) return;
    file = f;
    rows = rs; cols = cs;
    if (rows * cols < graph_num) {
      rows = int( sqrt(double(graph_num)) + 0.5 );
      cols = int( ceil( double(graph_num) / double(rows) ) );
    }
    double dx = 1.0 / double(cols);
    double dy = 1.0 / double(rows);  

    fprintf(file, "@ version 40000 \n");
    //fprintf(file, "@page size 700 595 \n");
    //fprintf(file, "@page scroll 5% \n");
    //fprintf(file, "@page inout 5% \n");
    fprintf(file, "@link page off \n");      

    for (int i = 0; i < rows; i++)
      {
      double y0 = 1.0 - (i + 1) * dy + dymargin / rows;
      double y1 = 1.0 - i * dy - dymargin / rows;
      for (int j = 0; j < cols; j++)
	{ 
        double x0 = j * dx + 1.5 * dxmargin / cols;
        double x1 = (j + 1) * dx - 0.5 * dxmargin / cols;
	int  graph = i * cols + j;
        if (graph >= graph_num) break;
        fprintf(file, "@with g%d\n", graph);
        fprintf(file, "@g%d on\n", graph);
        fprintf(file, "@g%d type xy\n", graph);

        fprintf(file, "@ view xmin %g\n", x0);
        fprintf(file, "@ view xmax %g\n", x1);
        fprintf(file, "@ view ymin %g\n", y0);
        fprintf(file, "@ view ymax %g\n", y1);
	}
      }
  }
		

//*******************************************************************************

 XmgrPlot::XmgrPlot(int sets, char is_log, const char *xlabel)
  {
  graph_id = graph_count;
 
  if ( ++graph_count > graph_num )  { 
     printf("XmgrPlot::XmgrPlot: too many graphs: [%d]>[%d]\n", graph_count, graph_num);
     exit(1); 
     }
  set_num = sets;
  set_count = 0;

  fprintf(file, "@with g%d\n", graph_id);
  fprintf(file, "@g%d on\n", graph_id);
  if (is_log) fprintf(file, "@g%d type logy\n", graph_id);

  // strcpy(x_label, xlabel);
  fprintf(file, "@ xaxis label \"%s\"\n", xlabel);
  fprintf(file, "@ xaxis label char size 1.00\n");
  fprintf(file, "@    legend off\n");

  for (int i = 0; i < set_num; i++)
    {
    fprintf(file, "@ s%d linestyle 1\n", i);
    fprintf(file, "@ s%d linewidth 2\n", i);
    if (i <= 11 ) fprintf(file, "@ s%d color %d\n", i, 11 - i);
    fprintf(file, "@ s%d symbol size 0.8\n", i);
    }
 
  if (set_num > 1)
    {
    double dx = 1.0 / double(cols);
    double dy = 1.0 / double(rows);  
    int row = graph_id / cols;
    int col = graph_id - row * cols;

    double ly = 1.0 - row * dy - dymargin / rows - dy_legend / rows;
    double lx = (col + 1) * dx - 0.5 * dxmargin / cols - dx_legend / cols;

    fprintf(file, "@    legend loctype view\n");
    fprintf(file, "@    legend layout 0\n");
    fprintf(file, "@    legend vgap 2\n");
    fprintf(file, "@    legend hgap 1\n");
    fprintf(file, "@    legend length 2\n");
    fprintf(file, "@    legend box on\n");
    fprintf(file, "@    legend box color 1\n");
    fprintf(file, "@    legend x1 %g\n", lx);
    fprintf(file, "@    legend y1 %g\n", ly);
    fprintf(file, "@    legend font 4\n");
    fprintf(file, "@    legend char size 1.000000\n");
    fprintf(file, "@    legend color 1\n");                                                      
    }
  }
  

//*******************************************************************************
//               C L A S S   X M G R   P O I N T   P L O T
//*******************************************************************************

 XmgrPointPlot::XmgrPointPlot(int sets, char is_log, double T, const char *xlabel) : 
 XmgrPlot(sets, is_log, xlabel), PlotObj(0, 0, " ", xlabel)
  {
  last = T;
  steps = XMGR_STEPS / cols;

  tptrs = new double *[sets];
  fptrs = new double *[sets];
  x_old = new double[sets];
  f_old = new double[sets];

  fmin = 1e20; fmax = -1e20;
  }
  
//*******************************************************************************

void XmgrPointPlot::get_range()
  {
  for (int i = 0; i < set_num; i++)
      {
      if ( *fptrs[i] < fmin )  fmin = *fptrs[i];
      if ( *fptrs[i] > fmax )  fmax = *fptrs[i];   
      }

   if ( (fabs((fmax - fmin)/(fmax + fmin + 1.0e-12)) < 1.0e-16) || (fmax <= fmin) ) 
     fmax = (fmin + 1e-8) * (1 + 1e-8);
  }
 
//*******************************************************************************

 void XmgrPointPlot::set_set(double *ptr, double *tptr, const char *txt)
    {
    fptrs[set_count] = ptr;
    tptrs[set_count] = tptr;
    x_old[set_count] = -100.0;
    f_old[set_count] = 0.0;
    strcat(win_title, txt);

    fprintf(file, "@with g%d\n", graph_id);
    fprintf(file, "@g%d on\n", graph_id);
    fprintf(file, "@ legend string %d \"%s\"\n", set_count, txt);

    set_count ++;

    if (set_count < set_num) strcat(win_title, ", ");
    else
      {
      first = *tptr;
      last += first;
      char format[30]; 
      double gx = format_double(last - first, 6, 1, format);

      fprintf(file, "@ world xmin %.10g\n", first);
      fprintf(file, "@ world xmax %.10g\n", last);
      fprintf(file, "@ xaxis tick major %g\n", gx);
      fprintf(file, "@ xaxis tick minor %g\n", gx/2.0);

      fprintf(file, "@ title \"%s vs. %s\"\n", win_title, x_label);
      fprintf(file, "@ title font 7\n");
      fprintf(file, "@ title size %g\n", 1.4 + 0.2 * (1 - cols) );
      fprintf(file, "@ title color %d\n", title_color);
      fprintf(file, "@ yaxis label \"%s\"\n", win_title);
      fprintf(file, "@ yaxis label char size 1.00\n");
      } 
    }

//*************************************************************************************
//                         X M G R   P L O T   D R A W
//*************************************************************************************

void  XmgrPointPlot::draw()
   {
   bool redrawFlag = true;
   int flag = 0;
   static double tscale = double( steps ) / last;

   for (int i = 0; i < set_num; i++)
     {
     // fprintf(stderr,"tscale=%g *tptrs[%d]=%g address=%ld\n",tscale,i,*tptrs[i],tptrs[i]);
     if ( int( *tptrs[i] * tscale ) == int( x_old[i] * tscale) ) redrawFlag = false;
     if ( !redrawFlag && ( fabs(f_old[i] - *fptrs[i]) < (fmax - fmin) * UPDATE_FRAC ) ) continue;
     x_old[i] = *tptrs[i];
     f_old[i] = *fptrs[i];
   
     if ( !flag ) 
       {
       get_range();
       if ( fmax - fmin < 2.0e-7 ) fmax = fmin + 2.0e-7;

       char   format[30];
       double dy = format_double(fmax - fmin, 6, 1, format);

       double y0 = dy * floor(fmin / dy); 
       double y1 = dy * ceil(fmax / dy); 
       if ( fabs( dy ) < 2.0e-7 ) dy = fmax - fmin;
       if ( fabs( y1 - y0 ) < 2.0e-7 ) y1 = y0 + dy;

       fprintf(file, "@WITH G%d\n", graph_id);
       fprintf(file, "@G%d ON\n", graph_id);
       
       fprintf(file, "@ world ymin %.10g\n", y0);
       fprintf(file, "@ world ymax %.10g\n", y1);
       fprintf(file, "@ yaxis tick major %g\n", dy);
       fprintf(file, "@ yaxis tick minor %g\n", dy*0.5);
       }
     flag = 1;

     fprintf(file, "@ g%d.s%d POINT %.8g, %.8g \n", graph_id, i, *tptrs[i], *fptrs[i] );   

     if ( (i + 1 == set_num) && (graph_id + 1 == graph_num) && redrawFlag)
       {
       fprintf(file, "@ redraw\n");
       fflush(file);
       }
      
     } // ** end loop over data sets
 }   

//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
//*******************************************************************************
//               C L A S S   1 D   X M G R   P I P E   P L O T
//*******************************************************************************

Xmgr1Dplot::Xmgr1Dplot(FieldObj *f, char islog, char dir, double coord1, double coord2, 
                       double T, int step_num, int m) : XmgrPlot(m, islog),
            FieldPlot1D(f, 0, dir, coord1, coord2)
{

  x_value = -100.0;
  last = T;
  steps = step_num;

  char format[30];
  double dx = format_double(coord[num-1] - coord[0], 6, 1, format);
    
  fprintf(file, "@ title \"%s\"\n", win_title);
  fprintf(file, "@ title font 7\n");
  fprintf(file, "@ title size %g\n", 1.4 + 0.2 * (1 - cols) );
  fprintf(file, "@ title color %d\n", title_color);
  fprintf(file, "@ yaxis label \"[%s] (\\8m\\4M)\" \n", field->ID);
  fprintf(file, "@ yaxis label char size 1.00\n");
  fprintf(file, "@ xaxis label \"distance (\\8m\\4m)\" \n");
  fprintf(file, "@ xaxis label char size 1.00\n");
  
  fprintf(file, "@ world xmin %g\n", coord[0]);
  fprintf(file, "@ world xmax %g\n", coord[num-1]);
  fprintf(file, "@ xaxis tick major %g\n", dx);
  fprintf(file, "@ xaxis tick minor %g\n", dx * 0.5);

  //set_colors();
}

//*************************************************************************************

void Xmgr1Dplot::get_range()
{
  fmin = get_value(field_index);
  fmax = get_value(field_index); 

  FieldPlot1D::get_range();
}

//*************************************************************************************
// xmgr colors: 0-white 1-black 2-red 3-green 4-blue 5-yellow 6-brown 7-gray 
// 8-violet 9-cyan 10-magenta 11-orange 12-indigo 13-maroon 14-turquise 15 -green4

void Xmgr1Dplot::set_colors()
{
  for (int i = set_num - 1; i >= 0; i--) {
    int c = int( 255.0 - i * 255.0 / double(set_num - 1) + 0.5 );
    fprintf(file, "@ cmap %d, %d, %d, %d \n", 29-i, 255, c, c);  
    fprintf(file, "@ s%d color %d \n", i, 29-i);  
    }
}

int Xmgr1Dplot::set_color(int n) 
{ 
  static int colors[30] = 
   {2, 2, 11, 5, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7}; 
  return colors[set_num-1-n]; 
}

 //*************************************************************************************
//                     1 D   X M G R   P L O T   D R A W
//*************************************************************************************

void  Xmgr1Dplot::draw()
   {
   double dt = last / double(steps);

   if  ( int( field->Time / dt ) == int( x_value / dt ) ) return;
   //fprintf(stderr,"last=%g steps=%d field->Time=%g x_value=%g dt=%g\n",last,steps,field->Time,x_value,dt);
   x_value = field->Time;

   fprintf(file, "@WITH G%d\n", graph_id);
   fprintf(file, "@G%d ON\n", graph_id);

   fprintf(file, "@ subtitle \"time = %gms\"\n", field->Time);   
   
   get_range();
   if ( fabs( fmax - fmin ) < 2.0e-7 ) fmax = fmin + 2.0e-7;
   char format[30];
   double dy = format_double(fmax - fmin, 6, 1, format);
   double y0 = dy * floor(fmin / dy); 
   double y1 = dy * ceil(fmax / dy); 
   if ( fabs( dy ) < 2.0e-7 ) dy = fmax - fmin;
   if ( fabs( y1 - y0 ) < 2.0e-7 ) y1 = y0 + dy;
   //printf("dy=%g fmin=%.10g fmax=%.10g y0=%.10g y1=%.10g\n",dy,fmin,fmax,y0,y1);

   fprintf(file, "@ world ymin %.10g\n", y0);
   fprintf(file, "@ world ymax %.10g\n", y1);
   fprintf(file, "@ yaxis tick major %g\n", dy);
   fprintf(file, "@ yaxis tick minor %g\n", dy * 0.5);   
 
   long ind = field_index;
   int set;

   for (set = 1; set < set_num; set++) 
     fprintf(file, "@ move g%d.s%d to g%d.s%d \n", graph_id, set, graph_id, set-1 );

   for (set = 0; set < set_num; set++) {
     fprintf(file, "@ s%d color %d \n", set, set_color(set) );  
     fprintf(file, "@ s%d linewidth 1\n", set);
     fprintf(file, "@ legend string %d \"t=%gms\" \n", set, field->Time - dt * (set_num - 1 - set) );  
   }
   fprintf(file, "@ s%d linewidth 2\n", set_num-1);
   
   for (int j = 0; j < num; j++, ind += incr)
     fprintf(file, "@ g%d.s%d POINT %g, %g\n", graph_id, set_num-1, coord[j], get_value(ind));

   if (graph_id + 1 == graph_num)
         {
	   //fprintf(file, "@ autoscale\n");
         fprintf(file, "@ redraw\n");
         fflush(file);
         }

 }   


//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
//*************************************************************************************
//                       U T I L I T Y   R O U T I N E S
//*************************************************************************************

double format_double(double number, int divisions, int sigdigits, char *fstr)
{
const  int maxzeros = 7;
double dx;
double mult = 1;
int    grid;
int    afterdec = 0;
int    beforedec = 0;

   strcpy(fstr,"");
   dx = number / double(divisions);

   if (dx < 1) 
     { 
     while ( dx * mult < 1 && dx * mult > 0.0 ) { mult *= 10; afterdec++; } 
     grid = int(dx * mult + 0.5);
     if ((grid & 1) && (grid != 1)  && (grid != 5)) grid--;  // prefer even grid
     dx = double(grid) / double(mult); 
     }
   else if (dx > 1)
     { 
     while (dx / mult >= 10) { mult *= 10; beforedec++; }
     grid = int(dx / mult + 0.5);
     if ((grid & 1) && (grid != 1)  && (grid != 5)) grid--;   // prefer even grid
     dx = double(grid) * double(mult); 
     }

   strcat(fstr,"%.");
   char sigdig[20];

   if (afterdec > maxzeros || beforedec > maxzeros) 
     {
     sprintf(sigdig, "%de", sigdigits);
     strcat(fstr, sigdig);
     }
   else
     if (afterdec > 0)                           // number < 1
       {
       sprintf(sigdig, "%df", afterdec + sigdigits - 1);
       strcat(fstr, sigdig);
       }
     else                                        // number >= 1
       {
       int digits = sigdigits - 1 - beforedec;
       if (digits < 0) digits = 0;
       sprintf(sigdig, "%df", digits);
       strcat(fstr, sigdig);
       }

  return dx;
}

//*************************************************************************************

void    MutePointPlot::draw()
   {
   double f = get_value();
   if (f > fmax) fmax = f; 
   else if (f < fmin) fmin = f;
   if ( *time < x_value && file != stdout ) {  // back step
     // fprintf(stdout, "oldtime=%g time=%g\n",x_value, *time);
     double x = 0, y = 0;
     fclose(file);
     FILE *temp = fopen(filename, "r");
     file = fopen("MutePointPlot.temp", "w");
     while (!feof(file)) {
       fscanf(temp,"%lf %lf",&x, &y);
       if (feof(temp) || x >= *time) break;
       x_value = x; fold = y;
       fprintf(file,"%.10g %.10g\n",x, y);  fflush(file);
       //       fprintf(stdout,"%.10g %.10g\n",x, y);
     }
     fclose(temp);
     rename("MutePointPlot.temp", filename);
   }
   else 
     if ( (int(*time * tscale) == int(x_value *tscale)) && (fabs(f - fold) < UPDATE_FRAC * (fmax - fmin)) ) 
       return;
 
   //       { 
   //	 if (x_value <=0 ) fprintf(stderr,"@@@ REDRAW FAILED!! time=%g tscale=%g\n",*time,tscale);
   //       return;
   //       }

   fold = f;
   x_value = *time; 
   fprintf(file, "%.10g %.10g \n", *time, get_value() );
   fflush(file);
   }

//*************************************************************************************

void format_print(double number, int sigdigits, char *strout)
{
char s[80];

format_double(number, 1, sigdigits, strout);
sprintf(s, strout, number);
strcpy(strout,s);
}


//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
//*************************************************************************************
//   T O K E N - S T R I N G   C O N S T R U C T O R   F O R   P L O T   A R R A Y
//*************************************************************************************
  
int PlotArray::get_plot_num(TokenString &TS)
{
  int graphs = 0;

  int mutes  = TS.token2_count("plot","mute") + TS.token2_count("plot","log_mute") +
               TS.token2_count("plot","mute_1D") + TS.token2_count("plot","log_mute_1D") +
               TS.token2_count("plot","dump") + TS.token_count("Export");
  int npoint = TS.token2_count("plot","point") + TS.token2_count("plot","log_point");
  int n1D    = TS.token2_count("plot","1D") + TS.token2_count("plot","log_1D");
  int n2D    = TS.token2_count("plot","2D") + TS.token2_count("plot","log_2D");
  int nraster= TS.token2_count("plot","raster") + TS.token2_count("plot","log_raster");

  char smethod[10];
  strcpy(smethod,"");

  if (TS.token_count("plot_print")) mutes += n1D + npoint;

  if (TS.token2_count("plot_method","xmgr"))  {
    int rows = 0, cols = 0;
    TS.trail2_pars("plot_method", "xmgr", 1, 'i', &rows, 'i', &cols);
    method = METHOD_XMGR;
    strcpy(smethod,"Xmgr");
    XmgrPlot::init(npoint + n1D, stdout, rows, cols);
    graphs = npoint + n1D;
  }
  else if (!TS.token2_count("plot_method","mute")) 
    {
    graphs = npoint + n1D + n2D + nraster;
    strcpy(smethod,"Ygl");
    method = METHOD_YGL;
    }
  else method = METHOD_MUTE;

  int total = graphs + mutes;

  if (VERBOSE) 
   fprintf(stderr,"\n### Setting up %d plots (%d %s graphs plus %d disc dumps):\n", 
           total, graphs, smethod, mutes);

  return  total;
}

//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
//*************************************************************************************

PlotArray::PlotArray(TokenString &params, FieldObj *Ca, BufferArray *Bufs, KineticObj *Gates)
{
  ygl_on = 0;
  count = 0; 
  plot_num = get_plot_num( params );

  if (!plot_num) fprintf(stderr,"PlotArray::PlotArray: no output (plot) commands\n"), exit(1);

  array   = new PlotObj *[plot_num];
  plotype = new char *[plot_num];

  if (!plot_num) return;

  char     ptype[100], prefix[128], filename[255], label[200];
  double   coord1, coord2, *time_ptr, *kin_ptr;
  FieldObj *field = 0;
  char     dir;
  
  strcpy(filename,"");
  strcpy(prefix,"temp");

  if ( params.token_count("plot_print") )  params.trail_pars("plot_print", 1, 's', prefix);

  int    dummy;
  double t = get_sim_time(params, dummy);
  int n = UPDATE_STEPS;

  int vcount = 0;  // counter for the verbose comment print

  //*************** Separate blosk for state export plot ****************************

  if (params.token_count("Export")) {
      double t0 = t, t1 = t, dt = t;
      plotype[count] = StrCpy("State dump");
      params.trail_pars("Export", 1, 'd', &t0, 'd', &t1, 'd', &dt, 's', filename);
      array[count++] = new StateDump(Gates, Ca, Bufs, t0, t1, dt, filename);
      if (VERBOSE)  fprintf(stderr," plot #1: State variable dump\n", count);
      vcount++;
    }

  //*********************************************************************************

  for (int i = 1; i <= params.token_count("plot"); i++)
    {
    char log_plot;
    int  hor = pref_width, ver = pref_height;

    params.trail_pars("plot", i, 's', ptype, 'E');
    if ( strncmp(ptype,"log",3) != 0 ) log_plot = 0; 
    else { log_plot = 1; strtok(ptype, "_"); strcpy( ptype, strtok('\0', " |") ); }

    int p = params.token_index("plot", i) + 2; 
    int set_num = 1;

    if ( method == METHOD_XMGR && 
          ( params.equal(p-1, "point") || params.equal(p-1, "log_point") ) ) 
    set_num = params.tokens_to_eol(p);

    if ( ( method == METHOD_XMGR ) && equal(ptype, "point") ) 
      {
      plotype[count] = StrCpy("point xmgr");
      array[count++] = new XmgrPointPlot(set_num, log_plot, t);
      } 

     //******************  MULTIPLE SETS IF XMGR POINT PLOT

    for (int set = 0; set < set_num; set++)
      { 
      strcpy(label, params[p + set]);
      kin_ptr = ResolveID(params, p + set, &time_ptr, Ca, Bufs, Gates);

      field = 0;
      if (Ca) {  if ( equal(label, Ca->ID) ) field = Ca; }
      if (Bufs) for (int i=0; i < Bufs->buf_num; i++)
        if ( equal(label, (Bufs->array[i])->ID) ) field = Bufs->array[i];

      if ( ( method == METHOD_XMGR ) && equal(ptype, "point") ) 
        ((XmgrPointPlot *)array[count-1])->set_set(kin_ptr, time_ptr, label);
      else if ( equal(ptype,"point") && (method == METHOD_YGL) ) 
        {
        ygl_on = 1;
        params.trail_pars("plot", i, 0,0, 0,0, 'i', &hor, 'i', &ver,  'E');
        plotype[count] = StrCpy("point");
        array[count++] = new PointPlot(kin_ptr, time_ptr, log_plot, t, n, label, "time=", hor, ver);
        }

      if ( (equal(ptype,"point") && params.token_count("plot_print") && (set == set_num - 1) ) 
           || equal(ptype,"mute") )
        {
	plotype[count] = StrCpy("point file");
        if ( equal(ptype,"mute") ) params.index_string(p+1, filename);
          else  sprintf(filename, "%s%s", prefix, label);
        array[count++] = new MutePointPlot(kin_ptr, time_ptr, log_plot, t, n, filename, label);
	}

      }  //@@@@@@@@@@@@@@@@@  END SET CYCLE @@@@@@@@@@@@@@@@@@@

    if ( equal(ptype, "point") || equal(ptype, "mute")) {  } 
       // hang around for the verbose comment

    else if ( equal(ptype,"2D") && field && (method == METHOD_YGL) ) {
      ygl_on = 1;
      plotype[count] = StrCpy("2D");
      params.trail_pars("plot", i, 0,0, 0,0, 'c', &dir, 'd', &coord1, 'i', &hor, 'i', &ver);
      array[count++] = new Plot2D(field, log_plot, dir, coord1, t, n, hor, ver);
      }
    else if ( equal(ptype,"dump") && field) {
      double t0 = t, t1 = t, dt = t;
      plotype[count] = StrCpy("file dump");
      params.trail_pars("plot", i, 0,0, 0,0, 'd', &t0, 'd', &t1, 'd', &dt, 's', filename);
      array[count++] = new FieldDump(field, t0, t1, dt, filename);
    }
    else if ( equal(ptype,"raster") && field && (method == METHOD_YGL) ) {
       ygl_on = 1;
       plotype[count] = StrCpy("raster");
       params.trail_pars("plot", i, 0,0, 0,0, 'c', &dir, 'd', &coord1, 'd', &coord2,
                        'i', &hor, 'i', &ver, 'E');
      array[count++] = new RasterPlot(field, log_plot, dir, coord1, coord2, t, n, hor, ver);
      }
    else if ( (equal(ptype,"1D") || equal(ptype,"mute_1D")) && field ) 
      {
      if ((method == METHOD_YGL) && equal(ptype,"1D")) {
        ygl_on = 1;
        plotype[count] = StrCpy("1D");
        params.trail_pars("plot", i, 0,0, 0,0, 'c', &dir, 'd', &coord1, 'd', &coord2, 'i', 
                        &hor, 'i', &ver,  'E');
        array[count++] = new Plot1D(field, log_plot, dir, coord1, coord2, t, n, hor, ver);
        }
      else if ((method == METHOD_XMGR) && equal(ptype,"1D")) {
        plotype[count] = StrCpy("1D xmgr");
        int nsteps = XMGR_STEPS, sets = 5;
        params.trail_pars("plot", i, 0,0, 0,0, 'c', &dir, 'd', &coord1, 'd', &coord2, 
                        'i', &nsteps, 'i', &sets, 'E');
        array[count++] = new Xmgr1Dplot(field, log_plot, dir, coord1, coord2, t, nsteps, sets);
      } 
      if ( params.token_count("plot_print") || equal(ptype,"mute_1D")) 
        {
        params.trail_pars("plot", i, 0,0, 0,0, 'c', &dir, 'd', &coord1, 'd', &coord2,  'E');
        char suffix[80];
        if (dir == 'x') sprintf(suffix,"-x-%.3g-%.3g",coord1,coord2);
          else if (dir == 'y') sprintf(suffix,"-y-%.3g-%.3g",coord1,coord2);
          else if (dir == 'z') sprintf(suffix,"-z-%.3g-%.3g",coord1,coord2);
	strcat(label,suffix);
        if ( !equal(ptype,"mute_1D") )  sprintf(filename, "%s%s", prefix, label);
          else params.index_string(p+4,filename);
        plotype[count] = StrCpy("1D file");
        array[count++] = new MutePlot1D(field, log_plot, dir, coord1, coord2, t, filename, label);
	}
      }
    else continue; // ignore some plots depending on plot_method

    if (VERBOSE) for(; vcount < count; vcount++) {
      char s[5];
      if (array[vcount]->log_plot) strcpy(s,"log "); else strcpy(s,"");
      fprintf(stderr," plot #%d: %s %splot of %s\n", vcount+1, plotype[vcount], s, label);
      }
    }
      // ****************   end loop over plot number
}

//*************************************************************************************
//*************************************************************************************
  
PlotArray::~PlotArray()  { 

  if (VERBOSE > 7) { fprintf(stderr,"> PlotArray: destructing %d plots: ", plot_num); 
                     fflush(stderr); }
   
  if (!count) return;

  tail_draw(); 

  for (int i = 0; i < count; i++)
    {
      //      if (method == METHOD_MUTE)

    if (equal(plotype[i],"point")           )  delete (PointPlot*)array[i];
    else if (equal(plotype[i],"raster")     )  delete (RasterPlot*)array[i];
    else if (equal(plotype[i],"1D")         )  delete (Plot1D*)array[i];
    else if (equal(plotype[i],"2D")         )  delete (Plot2D*)array[i];
    else if (equal(plotype[i],"point xmgr") ) delete (XmgrPointPlot*)array[i];
    else if (equal(plotype[i],"1D xmgr")    ) delete (Xmgr1Dplot*)array[i];
    else if (equal(plotype[i],"point file") ) delete (MutePointPlot*)array[i];
    else if (equal(plotype[i],"1D file")    ) delete (MutePlot1D*)array[i];
    else if (equal(plotype[i],"dump")       )  delete (FieldDump*)array[i];

    delete plotype[i];
    }

  delete [] array;
  delete [] plotype;

  if (VERBOSE > 7) fprintf(stderr,"done\n"); 
}

//*************************************************************************************
//*************************************************************************************

void Export(const char *filename, KineticObj *KO, FieldObj *Ca, BufferArray *BA)
{

  FILE *f;
 
  if (VERBOSE) 
    fprintf(stderr,"\n#### Dumping all state variables into file %s\n", filename);

  if (!(f = fopen(filename, "w")) )
    { 
    fprintf(stderr,">> Export: cannot open file %s\n",filename);
    exit(1);
    }


  if (Ca) {
    fwrite( (void *)&(Ca->size), sizeof(long), 1, f);
    fwrite( (void *)(Ca->elem), sizeof(double), Ca->size, f);
  }
  if (BA) 
     for (int i=0; i< BA->buf_num; i++)
        fwrite( (void *)(BA->array[i]->elem), sizeof(double), BA->array[i]->size, f);

  if (KO) {
     fwrite( (void *)&KO->var_num, sizeof(int), 1, f);
     fwrite( (void *)(KO->var->elem), sizeof(double), KO->var_num, f);
  }

  char endtag[7];
  strcpy(endtag,"endtag");
  fwrite( (void *)endtag, 7, 1, f);
}


//**************************************************************************************************

void Import(const char *filename, KineticObj *KO, FieldObj *Ca, BufferArray *BA)
{
  FILE *f;
 
  if (VERBOSE) 
    fprintf(stderr,"\n#### Importing all state variables from file %s\n", filename);

  if (!(f = fopen(filename, "r")) )
    { 
    fprintf(stderr,">> Import: cannot open file %s\n",filename);
    exit(1);
    }

  long n;
  long Size;

  if (Ca) {
    Size = Ca->size;
    fread( (void *)&n, sizeof(long), 1, f);
    if (n != Size)
      {
      fprintf(stderr,"\n>> Can't import the concentration fields: wrong number of elements (%ld vs %ld)\n", 
              n, Size);
      exit(1);
      }
    fread( (void *)(Ca->elem), sizeof(double), Size, f);
  }
  if (BA) 
     for (int i=0; i< BA->buf_num; i++)
        fread( (void *)(BA->array[i]->elem), sizeof(double), Size, f);

  if (KO) {
    int m;
    fread( (void *)&m, sizeof(int), 1, f);
    if (m != KO->var_num)
      {
      fprintf(stderr,"\n>> Can't import the kinetic variables: wrong number of elements (%ld vs %ld)\n", 
              m, KO->var_num);
      exit(1);
      }
    fread( (void *)(KO->var->elem), sizeof(double), KO->var_num, f);
    KO->locations->reset();
    KO->Evaluate();
  }

  char endtag[7];
  fread( (void *)endtag, 7, 1, f);
  if ( !equal(endtag,"endtag") )  {
    fprintf(stderr,"\n>> Import: file size mismatch\n");
    exit(1);
    }

}

//**************************************************************************************************
