
#ifndef GATE_H
#define GATE_H

#define  KINETIC_DEFAULT_EPS 1.0e-6
#define  DT_MIN 1.0e-14   // smallest time step

//*************************************************************************************

class KineticObj
{

 public:

  VectorObj      *var, *varOld;
  VectorObj      *ident;
  int            var_num;
  int            ident_num;
  char           **var_id;
  char           **ident_id;
  KinExpression  **formulas;

  TableArray     *tables;
  PeakTrackArray *facilitation;
  InterpolArray  *locations;
  
  double Time, TimeOld;

  KineticObj(int n, int m) : var_num(n), ident_num(m), Time(0.0) { initialize(); }
  KineticObj(TokenString &, FieldObj *Catalyst, BufferArray *B);
  void initialize();
  void Evaluate();

  VectorObj derivative();                      // returns dt satisfying accuracy eps
  double    RungeKuttaAdaptive(double T, double eps, class PlotArray *plots = 0, 
                               double total = 0.0, double dt = 0, long max_steps = 40000);   

  char      eq_flag;

  ~KineticObj();

  void saveState() { *varOld = *var; TimeOld = Time; 
                     tables->saveState(); facilitation->saveState(); locations->saveState(); }

  void recoverState() { *var = *varOld; Time = TimeOld; 
                     tables->recoverState(); facilitation->recoverState(); locations->recoverState(); }

};


//*************************************************************************************


#endif
