#ifndef GLOBAL_H
#define GLOBAL_H

int VERBOSE = 2;
int DIMENSIONALITY = 3;
int GEOMETRY = 0;

const int  SPHERICAL =  1;
const int  CYLINDRICAL = 2;

int     xsize=0, ysize=0, zsize=0;
long    Size, xysize;

double  *xgrid, *ygrid, *zgrid;
double  *xcoord, *ycoord, *zcoord;
double  *dxplus, *dxminus, *dxcenter;
double  *dyplus, *dyminus, *dycenter;
double  *dzplus, *dzminus, *dzcenter;
double  *dxinv,  *dyinv,   *dzinv;
double  *dx2inv, *dy2inv,  *dz2inv;

//*****************************************************************************
//                    G R I D   O P E R A T I O N S
//*****************************************************************************

   
 void set_dimensions(int nx, int ny, int nz);
 void set_dimensions(TokenString &params);

 void grid_stretch(double *grid, int N, double factor, int i0, int i1);
 void grid_stretch(char dir, double factor, double x1, double x2);
 void grid_stretch(TokenString &params);
 void grid_clamp(TokenString &params);

 void initScheme();
 void initSchemeSpherical();
 void initSchemeConical();

 void kill_grid();
 long   location_to_index(double x, double y, double z);
 int    location_to_index(double *, double *, int, double);

 double nearest_location(double *coord, double *grid, int size, double c)
   { return coord[ location_to_index(coord, grid, size, c) ]; }

 void split(long, int &, int &, int &);

 char in_bounds(double, double, double);


//*****************************************************************************
//     B O U N D A R Y   C O N D I T I O N   D A T A   S T R U C T U R E S
//*****************************************************************************

   double  *bc_deriv, *bc_lin, *bc_coef;
   char    **bc_id;
   int     bc_type_num;

   void  set_bc_types(TokenString &params);
   void  set_bc_types(int n);
   void  set_bc_type(double a, double b, double c, const char *id = "");
   int   bcidtoint(const char *);
   void  kill_bc();


#endif
