#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <stream.h>
#include "syntax.h"
#include "vector.h"
#include "box.h"
#include "field.h"
#include "expr.h"
#include "interpol.h"

extern  double *xcoord, *ycoord, *zcoord; 
extern  int    xsize, ysize, zsize;
extern  long   xysize; 
extern  int    VERBOSE;
extern  char   in_bounds(double, double, double);

extern int DIMENSIONALITY;
extern int GEOMETRY;
extern const int SPHERICAL;

//**************************************************************************

double InterpolObj::Evaluate(double t)
   {
     if (average) { 
        eigentime = *tptr; 
        if (GEOMETRY == 1) 
            { return result = Field->averageSpherical(); }
        else if (GEOMETRY == 2) 
            { return result = Field->averageConical(); }
        else return result = Field->average();
     }
   
     eigentime = t;

     if (*tptr > newtime)   // field has been updated since last interpolation evaluation
       {
       oldtime = newtime;
       oldres = newres;
       newtime = *tptr;
       newres  = *f000 * p000 + *f001 * p001 + *f010 * p010 + *f011 * p011 +   
                *f100 * p100 + *f101 * p101 + *f110 * p110 + *f111 * p111 ;
       }

     if (newtime > oldtime) // temporal extrapolation using values at two different times
       result = ( (newtime - t) * oldres + (t - oldtime) * newres ) / (newtime - oldtime);
     else result = newres;
     
     return result;
   }
//**************************************************************************

void InterpolObj::reset()
   {
   if (average) return;

   result = *f000 * p000 + *f001 * p001 + *f010 * p010 + *f011 * p011 +   
            *f100 * p100 + *f101 * p101 + *f110 * p110 + *f111 * p111 ;

   oldres = newres = result;
   oldtime = newtime = *tptr;   
   }

//*******************************************************************************
//*******************************************************************************

InterpolObj::InterpolObj(TokenString &Param, int p, FieldObj *FO)
   {
   char   xyz[255], brackets[200], *s;
   double x=0.0, y=0.0, z=0.0;

   // while( !strchr(xyz,')') ) strcat(xyz, Param[++p]);
   
   Field = FO;
   double *field = (*FO)();
   tptr = &(FO->Time); 
 
   strcpy(xyz, Param[p]);
   strcpy(brackets, FO->ID); strcat(brackets,"[]");

   if ( equal(brackets, xyz) )
     {
     ID = StrCpy( xyz ); 
     average = true;
     Evaluate(0.0);
     return;
     }

   average = false;

   s = strtok(xyz,"["); 

   if ( DIMENSIONALITY > 2 )
     {
     s = strtok('\0',",");
     if ( s && Param.is_number(s) ) x = Param.token_double(s); 
       else  fprintf(stderr,"InterpolObj::InterpolObj: can't parse [%s]\n",Param[p]);
     }

   if ( DIMENSIONALITY > 1 )
     {
     s = strtok('\0',",");
     if ( s && Param.is_number(s) ) y = Param.token_double(s); 
       else  fprintf(stderr,"InterpolObj::InterpolObj: can't parse [%s]\n",Param[p]);
     }

   s = strtok('\0',"]");
     if ( s && Param.is_number(s) ) z = Param.token_double(s); 
       else  fprintf(stderr,"InterpolObj::InterpolObj: can't parse [%s]\n",Param[p]);
 
   if ( DIMENSIONALITY == 1 ) { x = z; y = z = 0.0; sprintf(xyz,"%s[%g]", FO->ID, x); }
   else if ( DIMENSIONALITY == 2 ) { x = y; y = z; z = 0.0; sprintf(xyz,"%s[%g,%g]", FO->ID, x, y); }
   else sprintf(xyz,"%s[%g,%g,%g]", FO->ID, x, y, z);

   if ( !in_bounds(x, y, z) ) 
     fprintf(stderr,"InterpolObj::InterpolObj: location [%g,%g,%g] is out of bounds\n", x, y, z);
   
   ID = StrCpy( xyz );

   token_ptr_source = &(Param.token_ptr[p]);   // restore at destruction time
   token_ptr_target = Param.token_ptr[p];      //  
   Param.token_ptr[p] = ID;                    // *** hijack the token pointer

   double p1 = 1.0, p2 = 1.0, p3 = 1.0;
   int    ix = 0, iy = 0, iz = 0;  
   
   while (x >= xcoord[ix] && ix < (xsize - 1) ) ix++;
   if ( DIMENSIONALITY > 1 )  while (y >= ycoord[iy] && iy < (ysize - 1) ) iy++;
   if ( DIMENSIONALITY > 2 )  while (z >= zcoord[iz] && iz < (zsize - 1) ) iz++;

   // fprintf(stderr,"\n(%d,%d,%d)==(%g,%g,%g) point_type=%d\n",
   //        ix,iy,iz,xcoord[ix],ycoord[iy],zcoord[iz], FO->region->point_type(ix,iy,iz) );

   int opt = 0;
   do {    // find an interpolation cube close to (ix,iy,iz) with all vertices 
           // inside the volume
     if ( FO->region->point_type(ix,iy,iz) && FO->region->point_type(ix-1,iy,iz) &&
          FO->region->point_type(ix,iy-1,iz) && FO->region->point_type(ix,iy,iz-1) &&
          FO->region->point_type(ix-1,iy-1,iz) && FO->region->point_type(ix-1,iy,iz-1) &&
          FO->region->point_type(ix,iy-1,iz-1) && FO->region->point_type(ix-1,iy-1,iz-1) )
     break;
     else switch(opt) {

     case 0: ix++; break;      case 1: ix-=2; break;     case 2: iy++; break;     
     case 3: ix++; break;      case 4: ix++; break;      case 5: iy-=2; break;
     case 6: ix--; break;      case 7: ix--; break;      case 8: ix++; iy++; iz++; break;
     case 9: ix++; break;      case 10: ix-=2; break;    case 11: iy++; break;     
     case 12: ix++; break;     case 13: ix++; break;     case 14: iy-=2; break;
     case 15: ix--; break;     case 16: ix--; break;     case 17: ix++; iy++; iz-=2; break;
     case 18: ix++; break;     case 19: ix-=2; break;    case 20: iy++; break;     
     case 21: ix++; break;     case 22: ix++; break;     case 23: iy-=2; break;
     case 24: ix--; break;     case 25: ix--; break;     case 26: ix++; iy++; iz-=2; break;

     case 27: fprintf(stderr,"\n**** InterpolObj failure: check interpolation point!\n"); exit(1);
     }
   } while (++opt);

   /*
   fprintf(stderr," opt=%d (%d,%d,%d)\n", opt, ix, iy, iz);
   fprintf(stderr,"x=%g xcoord[ix-1]=%g xcoord[ix]=%g\n",x, xcoord[ix-1],xcoord[ix]);
   fprintf(stderr,"y=%g ycoord[iy-1]=%g ycoord[iy]=%g\n",y, ycoord[iy-1],xcoord[iy]);
   fprintf(stderr,"z=%g zcoord[iz-1]=%g zcoord[iz]=%g\n",z, zcoord[iz-1],xcoord[iz]);
   */

   p1 = (x - xcoord[ix-1]) / (xcoord[ix] - xcoord[ix-1]);
   if (DIMENSIONALITY > 1) p2 = (y - ycoord[iy-1]) / (ycoord[iy] - ycoord[iy-1]);
   if (DIMENSIONALITY > 2) p3 = (z - zcoord[iz-1]) / (zcoord[iz] - zcoord[iz-1]);

   f000 = f001 = f010 = f011 = f100 = f101 = f110 = f111 = field + ix + iy * xsize + iz * xysize;

   if ( ix > 0 ) { f000 -= 1;      f001 -= 1;      f010 -= 1 ;     f011 -= 1; }
   if ( iy > 0 ) { f000 -= xsize;  f001 -= xsize;  f100 -= xsize ; f101 -= xsize; }
   if ( iz > 0 ) { f000 -= xysize; f010 -= xysize; f100 -= xysize; f110 -= xysize ; }

   p111 =         p1  *      p2  *      p3 ;
   p000 =    (1 - p1) * (1 - p2) * (1 - p3);
   p100 =         p1  * (1 - p2) * (1 - p3);
   p010 =    (1 - p1) *      p2  * (1 - p3);
   p001 =    (1 - p1) * (1 - p2) *      p3 ;
   p110 =         p1  *      p2  * (1 - p3);
   p101 =         p1  * (1 - p2) *      p3 ;
   p011 =    (1 - p1) *      p2  *      p3 ;
 
   result = *f000 * p000 + *f001 * p001 + *f010 * p010 + *f011 * p011 +   
            *f100 * p100 + *f101 * p101 + *f110 * p110 + *f111 * p111 ;

   //fprintf(stderr,"p1=%g p2=%g p3=%g\n",p1,p2,p3);
   //fprintf(stderr,"(%ld,%g)(%ld,%g)(%ld,%g)(%ld,%g)(%ld,%g)(%ld,%g)(%ld,%g)(%ld,%g)\n",
   //   *f000,p000,*f001,p001,*f010,p010,*f011,p011,*f100,p100,*f101,p101,*f110,p110,*f111,p111);

   oldres = newres = result;
   oldtime = newtime = *tptr;   
}

//*******************************************************************************

InterpolArray::InterpolArray(TokenString &Param, FieldObj *Ca, BufferArray *Bufs)
{
  interpol_num = 0;

     if (Ca) { 
      char *w = new char[strlen( Ca->ID ) + 2];
      sprintf(w,"%s[",Ca->ID);
      interpol_num += Param.token_count_(w);
      delete [] w;
      }

    if (Bufs)
      for (int j = 0; j < Bufs->buf_num; j++)
	{
        char *w = new char[strlen( Bufs->array[j]->ID ) + 2];
        sprintf(w,"%s[",Bufs->array[j]->ID);
        interpol_num += Param.token_count_(w);
        delete [] w;
	}

    array = new InterpolObj *[interpol_num];
    int count = 0;

    if (!interpol_num) return; 

    if (VERBOSE) { fprintf(stderr,"\n### Interpolating: "); fflush(stderr); }

    if (Ca) { 
      char *w = new char[strlen( Ca->ID ) + 2];
      sprintf(w,"%s[",Ca->ID);
      for (int i = 0; i < Param.token_count_(w); i++)
        {
        int p0 = Param.token_index_(w, i+1);
        array[count] = new InterpolObj(Param, p0, Ca );
        if (VERBOSE) fprintf(stderr," %s",array[count]->ID);
        count ++;
        }
      delete [] w;
      }

    if (Bufs)
      for (int j = 0; j < Bufs->buf_num; j++)
	{
        char *w = new char[strlen( Bufs->array[j]->ID ) + 2];
        sprintf(w,"%s[",Bufs->array[j]->ID);
        for (int i = 0; i < Param.token_count_(w); i++)
          {
          int p0 = Param.token_index_(w, i+1);
          array[count] = new InterpolObj(Param, p0, Bufs->array[j] );
          if (VERBOSE) fprintf(stderr," %s",array[count]->ID);
          count ++;
          }
        delete [] w;
	}

    if (VERBOSE) fprintf(stderr,"\n  # Compressing interpolation array: ");

    InterpolObj *temp;

    for (int i = 0; i < interpol_num - 1; i++ )
      for  (int j = i + 1; j < interpol_num; j++ )
	  if ( equal(array[i]->ID, array[j]->ID) )
	    {
	    if (VERBOSE) { fprintf(stderr,"."); fflush(stderr); }
            interpol_num--;
	    temp = array[interpol_num]; 
            array[interpol_num] = array[j];
            array[j] = temp;
	    }

    if (VERBOSE) fprintf(stderr," done\n");
}

//*******************************************************************************

