
#ifndef INTERPOL_H
#define INTERPOL_H

//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
//*******************************************************************************


class InterpolObj
{
protected:

  FieldObj *Field;
  bool     average;  // true if InterpolObj simply tracks a field integral over volume

  double *f000, *f001, *f010, *f011, *f100, *f101, *f110, *f111;
  double p000, p001, p010, p011, p100, p101, p110, p111;

  double oldres, newres, oldtime, newtime;
  double oldres0, newres0, oldtime0, newtime0;

  char   **token_ptr_source, *token_ptr_target;  // restore 

public:

 double *tptr;
 double result;
 double eigentime;

 char   *ID;

 InterpolObj() { ID = 0; }
 
 /*
 InterpolObj(const InterpolObj &IO) 
   {
   ID = StrCpy(IO.ID);
   tptr = IO.tptr;
   token_ptr_source = IO.token_ptr_source;
   token_ptr_target = IO.token_ptr_target;
   *token_ptr_source = ID;
   f000 = IO.f000; f001 = IO.f001; f010 = IO.f010; f011 = IO.f011; 
   f100 = IO.f100; f101 = IO.f101; f110 = IO.f110; f111 = IO.f111; 
   p000 = IO.p000; p001 = IO.p001; p010 = IO.p010; p011 = IO.p011; 
   p100 = IO.p100; p101 = IO.p101; p110 = IO.p110; p111 = IO.p111; 
   }
 */

 InterpolObj(TokenString &Param, int p, FieldObj *);

 ~InterpolObj() { if (VERBOSE > 7) 
                     fprintf(stderr," > destructing InterpolObj %s ",ID), fflush(stderr);
                  if (!average) *token_ptr_source = token_ptr_target;
                  delete [] ID; 
                  if (VERBOSE > 7) fprintf(stderr," done\n");
                }

 double Evaluate(double t);
 void   reset();

 void saveState() { oldres0 = oldres; newres0 = newres; oldtime0 = oldtime; newtime0 = newtime; }
 void recoverState() { oldres = oldres0; newres = newres0; oldtime = oldtime0; newtime = newtime0;  }
};

//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
//*******************************************************************************

class InterpolArray
{
 public:

  InterpolObj  **array;
  int          interpol_num;

  InterpolArray(TokenString &, FieldObj *Ca, BufferArray *Bufs);

  ~InterpolArray()  { if (VERBOSE > 7) 
                        fprintf(stderr,"> destructing InterpolArray (%d elements) \n",interpol_num);
                      for (int i=0; i<interpol_num; i++) delete array[i]; 
                      delete [] array; 
                      if (VERBOSE > 7) fprintf(stderr,"> destructed InterpolArray\n");}  

  void Evaluate(double t)  {
    for (int i = 0; i < interpol_num; i++) array[i]->Evaluate(t);
  }

  void reset()  {  for (int i = 0; i < interpol_num; i++) array[i]->reset();  }

  void saveState() { for (int i = 0; i < interpol_num; i++) array[i]->saveState(); }
  void recoverState() { for (int i = 0; i < interpol_num; i++) array[i]->recoverState(); }
};

//**************************************************************************

#endif
