#ifndef PDE_ST_H
#define PDE_ST_H

double *right, *diag, *sup, *sub;

void    init_tridiag(int);
void    kill_tridiag();

typedef int (*CaMethod)(FieldObj &, VectorObj &, BufferArray &, double dt, double);
typedef int (*BufMethod)(BufferObj &, VectorObj &, FieldObj &, double dt, double); 

int Buf1Dstep(BufferObj &Buf, VectorObj &Buf_next, FieldObj &Ca, double dt, double);
int Ca1Dstep(FieldObj &Ca, VectorObj &Canext, BufferArray &, double dt, double);

int Buf2Dstep(BufferObj &Buf, VectorObj &Buf_next, FieldObj &Ca, double dt, double);
int Ca2Dstep(FieldObj &Ca, VectorObj &Canext, BufferArray &, double dt, double);

int Buf2DFIstep(BufferObj &Buf, VectorObj &Buf_next, FieldObj &Ca, double dt, double);
int Ca2DFIstep(FieldObj &Ca, VectorObj &Canext, BufferArray &, double dt, double);

int BufEu2Dstep(BufferObj &Buf, VectorObj &Buf_next, FieldObj &Ca, double dt, double);
int CaEu2Dstep(FieldObj &Ca, VectorObj &Canext, BufferArray &, double dt, double);

int BufEulerStep(BufferObj &Buf, VectorObj &Buf_next, FieldObj &Ca, double dt, double);
int CaEulerStep(FieldObj &Ca, VectorObj &Canext, BufferArray &, double dt, double);

int CaADIstep(FieldObj &Ca, VectorObj &Canext, BufferArray &, double dt, double);
int BufADIstep(BufferObj &Buf, VectorObj &Buf_next, FieldObj &Ca, double dt, double);

int CaLinADIstep(FieldObj &Ca, VectorObj &Canext, BufferArray &, double dt, double);
int BufLinADIstep(BufferObj &Buf, VectorObj &Buf_next, FieldObj &Ca, double dt, double);

int CaFIADIstep(FieldObj &Ca, VectorObj &Canext, BufferArray &, double dt, double);
int BufFIADIstep(BufferObj &Buf, VectorObj &Buf_next, FieldObj &Ca, double dt, double);

void FixedTimeStep(CaMethod, BufMethod, FieldObj &u, BufferArray &Bufs, KineticObj &Gates,
           double T, int n, class PlotArray &plots, double eps);


void getMethod(char *smethod, CaMethod &, BufMethod &);

void SetTimeStep(CaMethod CaStep, FieldObj &Ca, BufferArray &Buf, double &dt, double eps);

void Adaptive(CaMethod CaStep, BufMethod BufStep, FieldObj &Ca, BufferArray &Buf, KineticObj &Gate,
	      double T, double dt0, PlotArray &plots, double eps, double &dtmax,
              double &expand, int &n0, double &more, int &nmax);

void Run(TokenString &, FieldObj &, BufferArray &, KineticObj &Gate, class PlotArray &plots);

void RunODEs(TokenString &, KineticObj &, class PlotArray *plots);

double get_sim_time(TokenString &params, int &sim_steps);

#endif
