#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <stream.h>
#include "syntax.h"
#include "vector.h"
#include "box.h"
#include "field.h"
#include "expr.h"
#include "table.h"
#include "peak.h"
#include "interpol.h"
#include "gate.h"

extern double get_sim_time(TokenString &, int &);
extern int    VERBOSE;

//**************************************************************************

double PeakTrackObj::Evaluate()
   {
   if ( *tptr >= T1 && *tptr <= T2 )
     if ( *pointer > peak )  peak = *pointer;

   //   fprintf(stderr,"time=%g value=%g peak=%g\n",*tptr,*pointer,peak);

   return peak;
   }

//*******************************************************************************

PeakTrackObj::PeakTrackObj(TokenString &Param, int i)
   {
   peak = 1e-12;

   int p = Param.token_index(PEAK_LABEL, i+1) - 1;

   ID = Param.StrCpy(p);
   // pointer = ResolveID(Param, p + 2, &tptr, Ca, Bufs, KO);

   T1 = Param.index_double(p + 3);
   T2 = Param.index_double(p + 4);
   int dummy;
   double Tmax = get_sim_time(Param, dummy);

   if (T1 < 0) T1 = 0.0;
   if (T2 > Tmax) T2 = Tmax;

   if ( T1 >= T2 )
     fprintf(stderr,">>> PeakTrackObj::PeakTrackObj Error: zero interval length (%g..%g)\n",T1,T2),
       exit(1);

   if (VERBOSE)
       fprintf(stderr,"  #  %s = max of %s on [%g,%g] \n", ID, Param[p+2], T1, T2);
   }


//*******************************************************************************
//*******************************************************************************
/*
void PeakTrackArray::copy(const PeakTrackArray& PTA)  
{ 
    peak_num = PTA.peak_num;
    array = new PeakTrackObj *[peak_num];

    for (int i = 0; i < peak_num; i++)
      array[i] = new PeakTrackObj(*PTA.array[i]); 
}    
*/
//*******************************************************************************

PeakTrackArray::PeakTrackArray(TokenString &Param)
{
  peak_num = Param.token_count(PEAK_LABEL);

  array = new PeakTrackObj *[peak_num];

  if (peak_num && VERBOSE) 
       fprintf(stderr,"\n### Setting up %d peak tracking variable(s):\n", peak_num);

  for (int i = 0; i < peak_num; i++)  array[i] = new PeakTrackObj(Param, i);

}

//*******************************************************************************


void PeakTrackArray::set_pointers(TokenString &Param, FieldObj *Ca, BufferArray *Bufs, KineticObj *KO)
{
// if (num && VERBOSE) fprintf(stderr,"\n### Setting up %d peak tracking variable(s):\n", num);

   for (int i = 0; i < peak_num; i++ ) array[i]->set_pointer(Param, i, Ca, Bufs, KO);
}

//*******************************************************************************
