
#ifndef SYNTAX_H
#define SYNTAX_H

#define MAX_STORAGE_LENGTH  8000
#define STORAGE_INCREASE    2000

#define MAX_TOKEN_NUM       1000
#define TOKEN_NUM_INCREASE  200

#define LINE_CONTINUE_TOKEN "..."
#define COMMENT_TOKEN "%"
#define CARR_RET_TOKEN ";"

#define INPUT_TOKEN "input"

#define ODE_LABEL "->"
#define IDENT_LABEL "=="

//***************************************************************************

inline bool is_letter(const char c) {
  if ( (c >= 'a' && c <='z') || (c >= 'A' && c <='Z') || c == '_' ) return true;
  else return false; }

inline bool is_numeral(const char c) {
  if (c >= '0' && c <= '9') return true;  else return false; }

inline bool is_sign(const char c) {
  if (c == '+' || c == '-') return true;  else return false; }

inline bool is_short_token(const char c) {
  if ( strchr("~!@#$%^&*()=-+\\{}|;:/<>?",c) ) return true;
  else return false;
} 

inline bool is_delim(const char c) {
  if ( !is_letter(c) && !is_numeral(c) && !is_short_token(c) && !strchr(".[,]\"\'`",c) ) return true;
  else return false;
} 

inline bool is_number(const char *s)    // beware of extreme bootsrap!!! 
{
 bool flag = false;
 for (unsigned int i=0; i < strlen(s); i++) 
   if ( is_numeral(s[i]) ) flag = true;
   else if ( !is_sign(s[i]) && !(s[i] == '.') && !(s[i] == 'e') && !(s[i] == 'E') )  return false;

 return flag;
}

inline char equal(const char *s1, const char *s2) 
   { return ( (strcmp(s1, s2) == 0) && (strlen(s1) == strlen(s2)) ) ? 1 : 0; }

inline char equal_(const char *s1, const char *s2)  // returns "1" if s2 is a prefix of s1
{ return ( strncmp(s2, s1, strlen(s2)) == 0 ); }

inline char *StrCpy(const char *s) 
  { char *d = new char[strlen(s) + 1]; return strcpy(d, s); }

inline char *StrCpy(const char *s, int n) 
  { char *d = new char[n+1]; strncpy(d, s, n); d[n] = '\0'; return d; }

//***************************************************************************


class TokenString
{
  friend class LoopObj;
  friend class InterpolObj;

private:

 long length, max_length;
 char *storage;
 int  token_num, max_token_num;
 char **token_ptr;

public:

 char *carr_ret;

 int verbose;
  
 char   is_number(int p) {  return is_number(token_ptr[p]);  };
 char   is_number(const char *);

 char equal(int p, const char *s) { return ::equal(token_ptr[p], s); }
 char equal(const char *s, int p) { return equal(p, s); }
 char equal(int p1, int p2)       { return equal(token_ptr[p1], p2); }

 char equal_(int p, const char *s) { return ::equal_(token_ptr[p], s); }

 TokenString(const char *fname, const char *extra = 0);
 void parse(const char *fname, const char *extra = 0);

 TokenString(const TokenString &);

 ~TokenString()  { delete [] storage; delete [] token_ptr; }

 void getToken(char *&ptr0, char *&ptr1, char *, bool);
 void addToken(const char *token, int tlength);

 char *operator[](int index) const  { return token_ptr[index]; };
   
 void print(const char *fname);
 void print(FILE * = stderr);

 int token_count(const char *token);
 int token2_count(const char *token1, const char *token2);
 int token3_count(const char *token1, const char *token2, const char *token3);
 int token_count_(const char *token);

 int token_index(const char *token, int i = 1);
 int token2_index(const char *token1, const char *token2, int i = 1); 
 int token3_index(const char *token1, const char *token2, const char *token3, int i = 1); 
 int token_index_(const char *token, int i = 1);

 void trail_pars(const char *token, int m, ... );
 void trail2_pars(const char *token1, const char *token2, int m, ... );
 
 double trail_double(const char *token, int i = 1);
 double trail2_double(const char *token1, const char *token2, int i = 1);

 int trail_int(const char *token, int i = 1);
 int trail2_int(const char *token1, const char *token2, int i = 1);

 long trail_long(const char *token, int i = 1);
 long trail2_long(const char *token1, const char *token2, int i = 1);


 void trail_string(const char *token, char *s, int i = 1) 
                  { trail_pars(token, i, 's', s, 'E'); };
 void trail2_string(const char *token1, const char *token2, char *s, int i = 1) 
                  { trail2_pars(token1, token2, i, 's', s, 'E'); };
 char *trail_string(const char *token, int i = 1) 
                   { return index_string( token_index( token, i) + 1 ); };
 char *trail2_string(const char *token1, const char *token2, int i = 1) 
                   { return index_string( token2_index( token1, token2, i) + 1); };


 double index_double(int p)           { return token_double(token_ptr[p]); };
 int    index_int(int p)              { return token_int(token_ptr[p]); };
 long   index_long(int p)             { return token_long(token_ptr[p]); };
 void   index_string(int p, char *s)  { token_string(token_ptr[p], s); };
 char  *index_string(int p)           { return token_string(token_ptr[p]); };

 double token_double(const char *);
 int    token_int(const char *);
 long   token_long(const char *);
 void   token_string(const char *, char *);
 char  *token_string(const char *);


 int    tokens_to_eol(int p);

 void   append(const char *);

 char *StrCpy(int p) { return ::StrCpy(token_ptr[p]); } 
};

#endif
