#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <stream.h>
#include "syntax.h"
#include "table.h"

extern int VERBOSE;

//**************************************************************************

TableObj::TableObj(const char *fname, const char *id) : file_name(StrCpy(fname)),
  ID(StrCpy(id))
{
  if ( !(file = fopen(fname, "r")) )
    { printf("TableObj::TableObj: file [%s] not found\n", fname); exit(1); }

  if (VERBOSE) { fprintf(stderr,"  # table %s (file \"%s\"): ", ID, fname); fflush(stderr); }

  num = index = 0;
  double x, y;
  char s[256];

  while (1)
    {
    fgets(s, 255, file);
    if (strchr(s, '#') || strchr(s, ';') || strchr(s,'%') ) continue;
    sscanf(s,"%lf %lf",&x, &y);    
    if (feof(file)) break;
    num ++;
    }

  xarray = new double[num];
  yarray = new double[num];

  rewind(file);
  num = 0;
  while (1)
    {
    fgets(s, 255, file);
    if (strchr(s, '#') || strchr(s, ';') || strchr(s,'%') ) continue;
    sscanf(s,"%lf %lf",&x, &y);    
    if (feof(file)) break;
    if (num) if (x <= xarray[num-1]) continue;
    xarray[num] = x; yarray[num] = y;
    num ++;
    }

  if (VERBOSE) fprintf(stderr,"%ld entries: first = (%g, %g), last = (%g, %g)\n",
	        num, xarray[0], yarray[0], xarray[num-1], yarray[num-1]);
  fclose(file);
}

//**************************************************************************
/*
void TableObj::copy(const TableObj& TO)
{
  num = TO.num; index = TO.index;

  ID = StrCpy(TO.ID); file_name = StrCpy(TO.file_name);

  xarray = new double[num];
  yarray = new double[num];

  for (int i = 0; i < num; i++) {
    xarray[i] = TO.xarray[i];
    yarray[i] = TO.yarray[i];
  }
}
*/
//**************************************************************************


double TableObj::Evaluate(double x)
{
  //  printf("table time = %g\n", x);

  while ( (xarray[index] > x) && (index > 0) ) index--;
  if (index < num - 2) while ( (x > xarray[index+1]) && (index < num - 2) ) index++;

  value = ( yarray[index] * (xarray[index+1] - x) + yarray[index+1] * (x - xarray[index]) ) / 
      ( xarray[index+1] - xarray[index] );
  /*
  fprintf(stderr,"TableObj::Evaluate: %g %g index=%ld num=%ld last:(%g,%g) next:(%g,%g) \n",
          x, value,index,num,xarray[num-1],yarray[num-1],xarray[num-2],yarray[num-2]);
  */
  return value;
}


//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
//**************************************************************************

  TableArray::TableArray(TokenString &Param) : table_num(Param.token_count(TABLE_LABEL))
  {
    array = new TableObj *[table_num];

    if (table_num && VERBOSE) fprintf(stderr,"\n### Setting up %d tables:\n", table_num);

    for (int i = 0; i < table_num; i++)
      {
      int p0 = Param.token_index(TABLE_LABEL, i+1) + 1;
      char *fname = Param.index_string(p0);
      array[i] = new TableObj(fname, Param[p0-2]);
      delete [] fname;
      }
  }

//**************************************************************************
/*
  void TableArray::copy(const TableArray& TA) 
  {
    table_num = TA.table_num;
    array = new TableObj *[table_num];

    for (int i = 0; i < table_num; i++)   *array[i] = *TA.array[i];
  }
 */
//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
//**************************************************************************
